import logging

from django.db import migrations

logger = logging.getLogger(__name__)


def get_group_mappings() -> dict:
    """Returns the mapping table for updating existing groups."""
    return {
        "Platform": ("Platform", 8),
        "Visitors": ("Marketplace", 1),
        "Specialist Offer (Catalog/Seller)": ("Catalog", 4),
        "Operations": ("Operations", 6),
        "Money": ("Payment Operations", 5),
        "Security": ("Security", 10),
        "Corporate IT": ("Corporate IT", 11),
        "Other": ("Other", 12),
        "Data": ("Data", 9),
    }


def get_new_groups() -> dict:
    """Returns a dictionary of new groups to be created."""
    return {
        "Marketing & Communication": 2,
        "Seller": 3,
        "Finance": 7,
    }


def add_new_groups(apps, _schema_editor):
    Group = apps.get_model("incidents", "Group")
    new_groups = get_new_groups()

    for name, position in new_groups.items():
        try:
            logger.info(f"Creating new group: '{name}' with order {position}")
            new_group = Group(name=name, order=position)
            new_group.save()
        except Exception:
            logger.exception(f"Failed to create new group '{name}'.")


def remove_new_groups(apps, _schema_editor):
    Group = apps.get_model("incidents", "Group")
    new_group_names = get_new_groups().keys()

    for name in new_group_names:
        try:
            logger.info(f"Removing group: '{name}'")
            group = Group.objects.get(name=name)
            group.delete()
        except Exception:
            logger.exception(f"Group '{name}' does not exist, skipping removal.")


def update_groups(apps, _schema_editor):
    Group = apps.get_model("incidents", "Group")
    group_mappings = get_group_mappings()

    updated_count = 0

    for old_name, (new_name, position) in group_mappings.items():
        try:
            logger.info(f"Updating group: '{old_name}' to '{new_name}'")
            group = Group.objects.get(name=old_name)
            group.name = new_name
            group.order = position
            group.save()
            updated_count += 1
        except Exception:
            logger.exception(f"Group '{old_name}' does not exist, cannot proceed with updates.")


def revert_group_names(apps, _schema_editor):
    Group = apps.get_model("incidents", "Group")
    reverse_mappings = {new_name: old_name for old_name, (new_name, _) in get_group_mappings().items()}

    updated_count = 0

    for new_name, old_name in reverse_mappings.items():
        try:
            group = Group.objects.get(name=new_name)
            logger.info(f"Restoring group '{new_name}' back to '{old_name}'")
            group.name = old_name
            group.save()
            updated_count += 1
        except Exception:
            logger.exception(f"Group '{new_name}' does not exist, skipping restoration.")


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0005_enable_from_p1_to_p5_priority"),
    ]

    operations = [
        migrations.RunPython(add_new_groups, remove_new_groups),
        migrations.RunPython(update_groups, revert_group_names),
    ]



from __future__ import annotations

import inspect
import logging
import operator
import re
from functools import reduce
from typing import TYPE_CHECKING, Any, TypeVar

from django.forms import Form
from slack_bolt.request.payload_utils import (
    to_command,
    to_event,
    to_message,
    to_options,
    to_shortcut,
    to_step,
    to_view,
)

from firefighter.slack.models.user import SlackUser
from firefighter.slack.slack_app import SlackApp
from firefighter.slack.slack_incident_context import (
    get_incident_from_context,
    get_user_from_context,
)
from firefighter.slack.utils import get_slack_user_id_from_body, respond
from firefighter.slack.views.modals.base_modal.form_utils import (
    SlackForm,
    slack_view_submission_to_dict,
)
from firefighter.slack.views.modals.base_modal.modal_utils import (
    open_modal,
    push_modal,
    update_modal,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from django.core.exceptions import ValidationError
    from slack_bolt.context.ack.ack import Ack
    from slack_bolt.request.request import BoltRequest
    from slack_bolt.response.response import BoltResponse
    from slack_sdk.models.blocks.blocks import Block
    from slack_sdk.models.views import View

T = TypeVar("T", bound=Form)
app = SlackApp()

logger = logging.getLogger(__name__)


class SlackModal:
    """Two main responsibilities:
    - Register the modal on Slack to open it with shortcuts or actions
    - Provide useful context for `build_modal_fn` and `handle_modal_fn`, such as body, incident, user, etc.
    """

    callback_id: str | re.Pattern[str]
    """Callback ID for the Slack View"""

    open_shortcut: str | None = None
    """[Slack shortcut](https://api.slack.com/interactivity/shortcuts) to open the modal."""

    open_action: str | list[str] | None = None
    update_action: str | list[str] | None = None
    push_action: str | list[str] | None = None

    build_modal_fn: Callable[..., View | list[Block]]
    handle_modal_fn: Callable[[Any], Any] | None

    def __init__(self) -> None:
        if self.handle_modal_fn is not None:
            self.handler_fn_args = inspect.getfullargspec(self.handle_modal_fn).args
            if len(self.handler_fn_args) > 0 and self.handler_fn_args[0] in {
                "self",
                "cls",
            }:
                self.handler_fn_args.pop(0)

        self.builder_fn_args = inspect.getfullargspec(self.build_modal_fn).args
        if len(self.builder_fn_args) > 0 and self.builder_fn_args[0] in {"self", "cls"}:
            self.builder_fn_args.pop(0)

        self._register_actions_shortcuts()

    def open_modal_aio(self, ack: Ack, body: dict[str, Any], **kwargs: Any) -> None:
        ack()

        view = self.build_modal_with_context(body, **kwargs)

        open_modal(view=view, body=body)

    def build_modal_with_context(self, body: dict[str, Any], **kwargs: Any) -> View:
        self._get_kwargs_builder(body, kwargs, self.builder_fn_args)
        if "body" in self.builder_fn_args:
            kwargs["body"] = body
        elif "body" in kwargs:
            del kwargs["body"]

        return self.build_modal_fn(**kwargs)  # type: ignore[return-value]

    def _register_actions_shortcuts(self) -> None:
        """Register the different actions and shortcuts, from the class attributes."""
        self._register_meta(
            app.action, self.open_action, self._handle_shortcut_action_open
        )
        self._register_meta(app.action, self.update_action, self._handle_action_update)
        self._register_meta(app.action, self.push_action, self._handle_action_push)
        self._register_meta(
            app.shortcut, self.open_shortcut, self._handle_shortcut_action_open
        )

        target = app.action if hasattr(self, "callback_action") else app.view
        self._register_meta(target, self.callback_id, self._handle_modal)

    def _handle_shortcut_action(
        self, request: BoltRequest, response: BoltResponse, **kwargs: Any
    ) -> tuple[View, dict[str, Any]]:
        fn_kwargs, body = self._get_handler_kwargs(
            request, response, self.builder_fn_args
        )
        if "body" not in fn_kwargs:
            fn_kwargs["body"] = body

        request.context.ack()
        return self.build_modal_with_context(**fn_kwargs), body

    def _handle_shortcut_action_open(
        self, request: BoltRequest, response: BoltResponse, **kwargs: Any
    ) -> None:
        view, body = self._handle_shortcut_action(request, response, **kwargs)
        open_modal(view=view, body=body)

    def _handle_action_update(
        self, request: BoltRequest, response: BoltResponse, **kwargs: Any
    ) -> None:
        view, body = self._handle_shortcut_action(request, response, **kwargs)
        update_modal(view=view, body=body)

    def _handle_action_push(
        self, request: BoltRequest, response: BoltResponse, *args: Any, **kwargs: Any
    ) -> None:
        view, body = self._handle_shortcut_action(request, response, **kwargs)
        push_modal(view=view, body=body)

    def _handle_modal(
        self, request: BoltRequest, response: BoltResponse, **kwargs: Any
    ) -> None:
        """Wrapper to add the required args to the handle_modal_fn."""
        if self.handle_modal_fn is None:
            raise ValueError("No handle_modal_fn defined!")

        fn_kwargs, body = self._get_handler_kwargs(
            request, response, self.handler_fn_args
        )

        if "incident" in self.handler_fn_args and fn_kwargs["incident"] is None:
            logger.error(f"Could not get the incident from slack context! Body: {body}")
            respond(
                body=body,
                text="Unexpected error when fetching this incident. Please tell @pulse (#tech-pe-pulse) about the context of this error.",
            )
            return
        if "user" in self.handler_fn_args and fn_kwargs["user"] is None:
            logger.error(f"Could not get the user from slack context! Body: {body}")
            respond(
                body=body,
                text="Unexpected error when fetching your user. Please tell @pulse (#tech-pe-pulse) about the context of this error.",
            )
            return

        self.handle_modal_fn(**fn_kwargs)  # type: ignore

    @staticmethod
    def _get_kwargs_builder(
        body: dict[str, Any], kwargs: dict[str, Any], args: list[str]
    ) -> None:
        """Adds `user` and `incident` to the kwargs from the body context, if they are needed and not present."""
        if "incident" in args and kwargs.get("incident") is None:
            kwargs["incident"] = get_incident_from_context(body)

        if "user" in args and kwargs.get("user") is None:
            sender_id = get_slack_user_id_from_body(body)
            if sender_id:
                kwargs["user"] = SlackUser.objects.get_user_by_slack_id(
                    slack_id=sender_id
                )
            else:
                logger.warning(
                    f"Could not get the user from slack context! Body: {body}"
                )
                kwargs["user"] = None

    @staticmethod
    def _get_handler_kwargs(
        request: BoltRequest, response: BoltResponse, args: list[str]
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """Adapted from Slack Bolt. Provide the kwargs for the function, from the Slack request and response.
        Additionally, it can provide the incident and user from the Slack context.

        Args:
            request (BoltRequest): The Slack request
            response (BoltResponse): The Slack response
            args (list[str]): List of args of the function

        Returns:
            tuple[dict[str, Any], dict[str, Any]]: kwargs for the function, body of the response
        """
        all_available_args = {
            "logger": logger,
            "client": request.context.client,
            "req": request,
            "request": request,
            "resp": response,
            "response": response,
            "context": request.context,
            # payload
            "body": request.body,
            "options": to_options(request.body),
            "shortcut": to_shortcut(request.body),
            "view": to_view(request.body),
            "command": to_command(request.body),
            "event": to_event(request.body),
            "message": to_message(request.body),
            "step": to_step(request.body),
            # utilities
            "ack": request.context.ack,
            "say": request.context.say,
            "respond": request.context.respond,
            # middleware
        }

        # Add Slack Bolt Context
        fn_kwargs: dict[str, Any] = {
            k: v for k, v in all_available_args.items() if k in args
        }
        body: dict[str, Any] = all_available_args["body"]  # type: ignore

        # Add custom context if needed
        if "incident" in args:
            fn_kwargs["incident"] = get_incident_from_context(body)

        if "user" in args:
            fn_kwargs["user"] = get_user_from_context(body)

        return fn_kwargs, body

    @staticmethod
    def _register_meta(
        fn_register: Callable[
            [str | re.Pattern[str]], Callable[[Callable[..., Any]], Any]
        ],
        attributes: list[str] | str | re.Pattern[str] | None,
        fn: Callable[..., Any] | None,
    ) -> None:
        if fn is None:
            logger.debug(
                f"Skipping registration {fn_register.__name__} with {attributes}"
            )
            return
        if attributes is not None:
            if isinstance(attributes, str | re.Pattern):
                fn_register(attributes)(fn)
            else:
                for action in attributes:
                    fn_register(action)(fn)


class ModalForm[T: Form](SlackModal):
    """Specific SlackModal to handle a Django form.

    Provided a Django form in form_class, it will handle:
    - generation of Slack Blocks
    - validation and submission of the form
    """

    form_class: type[T] | None
    form: T

    def get_form_class(self) -> SlackForm[T]:
        if self.form_class is None:
            raise ValueError("No form_class defined!")
        return SlackForm(self.form_class)

    def handle_form_errors(
        self,
        ack: Ack,
        body: dict[str, Any],
        forms_kwargs: dict[str, Any],
        *,
        ack_on_success: bool = True,
    ) -> SlackForm[T] | None:
        slack_form = self.get_form_class()(
            data=slack_view_submission_to_dict(body), **forms_kwargs
        )
        form = slack_form.form
        if form.is_valid():
            if ack_on_success:
                ack()
        else:
            ack(
                response_action="errors",
                errors={
                    k: self._concat_validation_errors_msg(v)
                    for k, v in form.errors.as_data().items()
                },
            )
            return None
        return slack_form

    @staticmethod
    def _concat_validation_errors_msg(err: list[ValidationError]) -> str:
        """Return a string of all validation errors."""
        return reduce(operator.add, ["; ".join(e.messages) for e in err], "")


class MessageForm[T: Form](SlackModal):
    """Form wrapper to use a Django form in a Slack message.

    Provided a Django form in form_class, it will handle:
    - generation of Slack Blocks
    - validation and submission of the form
    """

    form_class: type[T] | None
    form: T

    def get_form_class(self) -> SlackForm[T]:
        if self.form_class is None:
            raise ValueError("No form_class defined!")
        return SlackForm(self.form_class, {})

    def handle_form_errors(
        self, ack: Ack, body: dict[str, Any], forms_kwargs: dict[str, Any]
    ) -> SlackForm[T] | None:
        slack_form = self.get_form_class()(
            data=slack_view_submission_to_dict(body), **forms_kwargs
        )
        form = slack_form.form
        # Only change one field, so only one data. We add the rest from the initial data.
        form.data = {**form.initial, **form.data}
        if form.is_valid():
            ack(text="Updated successfully!", response_type="ephemeral")
        else:
            logger.warning("Form errors in Slack context.")
            logger.warning(form.errors.as_data())

            ack(text="Error", response_type="ephemeral")

        return slack_form



from __future__ import annotations

from django.apps import AppConfig


class IncidentsConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "firefighter.incidents"
    label = "incidents"

    def ready(self) -> None:
        from firefighter.incidents import observability, tasks
        from firefighter.incidents.models.incident_update import set_event_ts



from __future__ import annotations

import logging
from copy import deepcopy
from unittest.mock import MagicMock, PropertyMock

import pytest
from django.apps import apps
from pytest_mock import MockerFixture

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.factories import IncidentFactory, UserFactory
from firefighter.incidents.models.incident import Incident
from firefighter.slack.views import CloseModal

logger = logging.getLogger(__name__)

if not apps.is_installed("firefighter.confluence"):
    pytest.skip("Confluence app not installed; skipping close modal tests", allow_module_level=True)


@pytest.mark.django_db
class TestCloseModal:
    @staticmethod
    def test_close_modal_build(mocker: MockerFixture) -> None:
        # Arrange
        modal = CloseModal()
        mocker.patch.object(
            Incident,
            "can_be_closed",
            new_callable=PropertyMock(return_value=(True, [])),
        )
        incident = IncidentFactory.build()
        incident.status = IncidentStatus.MITIGATED

        # Act
        res = modal.build_modal_fn(incident=incident, body={})

        # Assert
        assert incident.can_be_closed[0] is True
        # To dict will validate the JSON as well (e.g. check the view title length)
        # Any validation error will raise an exception
        assert res.to_dict()

        values = res.to_dict()
        assert "blocks" in values
        assert len(values["blocks"]) == 7

    @staticmethod
    def test_close_modal_build_initial_values(
        mocker: MockerFixture, incident: Incident
    ) -> None:
        """Test that the initial values are set correctly from the incident."""
        # Arrange
        modal = CloseModal()
        mocker.patch(
            "firefighter.incidents.models.incident.Incident.can_be_closed",
            return_value=(True, []),
            new_callable=mocker.PropertyMock,
        )
        incident.status = IncidentStatus.MITIGATED
        incident.title = "This is the title"
        incident.description = "This is the description"

        # Act
        res = modal.build_modal_fn(incident=incident, body={})
        values = res.to_dict()

        # Assert
        assert values["blocks"][1]["element"]["initial_value"] == "This is the title"
        assert (
            values["blocks"][2]["element"]["initial_value"] == "This is the description"
        )

    @staticmethod
    def test_close_modal_build_cant_close(incident: Incident) -> None:
        # Arrange
        modal = CloseModal()
        # Use MITIGATING (Mitigating) status - cannot close from this status without going through MITIGATED
        incident.status = IncidentStatus.MITIGATING

        # Act
        res = modal.build_modal_fn(incident=incident, body={})

        # Assert

        # To dict will validate the JSON as well (e.g. check the view title length)
        # Any validation error will raise an exception
        assert res.to_dict()

        values = res.to_dict()
        assert "blocks" in values
        assert len(values["blocks"]) >= 5
        assert (
            "This incident can't be closed yet." in values["blocks"][0]["text"]["text"]
        )

    @staticmethod
    def test_close_modal_build_shows_mitigated_status_requirement(
        mocker: MockerFixture, incident: Incident
    ) -> None:
        """Test that the modal shows STATUS_NOT_MITIGATED error with proper references.

        This test covers lines 151, 154, 159 in close.py where MITIGATED.label is used.
        """
        # Arrange
        modal = CloseModal()
        incident.status = IncidentStatus.INVESTIGATING

        # Mock requires_closure_reason to return False so we bypass the closure reason modal
        mocker.patch(
            "firefighter.slack.views.modals.utils.UpdateStatusForm.requires_closure_reason",
            return_value=False
        )

        # Mock can_be_closed to return False with STATUS_NOT_MITIGATED reason
        mocker.patch.object(
            Incident,
            "can_be_closed",
            new_callable=PropertyMock(return_value=(False, [("STATUS_NOT_MITIGATED", "Status not mitigated")])),
        )

        # Act
        res = modal.build_modal_fn(incident=incident, body={})

        # Assert
        assert res.to_dict()
        values = res.to_dict()
        assert "blocks" in values

        # Convert blocks to text for easier searching
        blocks_text = str(values["blocks"])

        # Verify that "Mitigated" (the label) appears in the error message
        # Line 154: text=f":warning: *Status is not _{IncidentStatus.MITIGATED.label}_* :warning:\n"
        # Line 159: text=f"You can only close an incident when its status is _{IncidentStatus.MITIGATED.label}_ or _{IncidentStatus.POST_MORTEM.label}_..."
        assert "Mitigated" in blocks_text
        assert "Post-mortem" in blocks_text or "Post Mortem" in blocks_text

        # Verify the error message structure
        assert "This incident can't be closed yet." in values["blocks"][0]["text"]["text"]

    @staticmethod
    def test_close_modal_build_shows_closure_reason_from_open(incident: Incident) -> None:
        # Arrange
        modal = CloseModal()
        incident.status = IncidentStatus.OPEN

        # Act
        res = modal.build_modal_fn(incident=incident, body={})

        # Assert
        assert res.to_dict()
        values = res.to_dict()
        assert "blocks" in values
        # Should show closure reason form
        assert "Closure Reason Required" in values["blocks"][0]["text"]["text"]

    @staticmethod
    def test_submit_empty_bodied_form() -> None:
        modal = CloseModal()
        ack = MagicMock()
        user = UserFactory.build()
        with pytest.raises(TypeError, match="Expected a values dict in the body"):
            modal.handle_modal_fn(ack=ack, body={}, incident=MagicMock(), user=user)

    @staticmethod
    def test_submit_valid_form_open_incident(
        mocker: MockerFixture, incident: Incident
    ) -> None:
        incident.status = IncidentStatus.OPEN

        modal = CloseModal()
        trigger_incident_workflow = mocker.patch.object(
            modal, "_trigger_incident_workflow"
        )
        respond = mocker.patch(
            "firefighter.slack.views.modals.close.respond", return_value=None
        )

        ack = MagicMock()
        user = UserFactory.build()
        user.save()

        modal.handle_modal_fn(
            ack=ack, body=valid_submission, user=user, incident=incident
        )

        # Assert
        ack.assert_called_once_with()
        trigger_incident_workflow.assert_not_called()
        respond.assert_called_once()

    @staticmethod
    def test_submit_valid_form_closable_incident(
        mocker: MockerFixture, incident: Incident
    ) -> None:
        modal = CloseModal()
        mocker.patch(
            "firefighter.incidents.models.incident.Incident.can_be_closed",
            return_value=(True, []),
            new_callable=mocker.PropertyMock,
        )
        incident.status = IncidentStatus.POST_MORTEM
        trigger_incident_workflow = mocker.patch.object(
            modal, "_trigger_incident_workflow"
        )

        ack = MagicMock()
        user = UserFactory.build()
        user.save()

        modal.handle_modal_fn(
            ack=ack, body=valid_submission, user=user, incident=incident
        )

        # Assert
        ack.assert_called_once_with()
        trigger_incident_workflow.assert_called_once()


valid_submission = {
    "type": "view_submission",
    "team": {"id": "T01FJ0NNFQD", "domain": "team-firefighter"},
    "user": {
        "id": "U020BFXLSKX",
        "username": "gabriel.doogabe",
        "name": "gabriel.doogabe",
        "team_id": "T01FJ0NNFQD",
    },
    "api_app_id": "A020BG97EG5",
    "token": "fake_",
    "trigger_id": "3914397216230.1528022763829.149fcfbe4712c05a4fb707c9792c2011",
    "view": {
        "id": "V03T2SMMWBV",
        "team_id": "T01FJ0NNFQD",
        "type": "modal",
        "blocks": [
            {
                "type": "input",
                "block_id": "title",
                "label": {
                    "type": "plain_text",
                    "text": "What's going on?",
                    "emoji": True,
                },
                "optional": False,
                "dispatch_action": False,
                "element": {
                    "type": "plain_text_input",
                    "action_id": "title",
                    "placeholder": {
                        "type": "plain_text",
                        "text": "Short, punchy description of what's happening.",
                        "emoji": True,
                    },
                    "multiline": False,
                    "min_length": 10,
                    "max_length": 128,
                    "dispatch_action_config": {
                        "trigger_actions_on": ["on_enter_pressed"]
                    },
                },
            },
            {
                "type": "input",
                "block_id": "description",
                "label": {"type": "plain_text", "text": "Summary", "emoji": True},
                "optional": False,
                "dispatch_action": False,
                "element": {
                    "type": "plain_text_input",
                    "action_id": "description",
                    "placeholder": {
                        "type": "plain_text",
                        "text": "Help people responding to the incident. This will be posted to #tech-incidents and on our internal status page.\nThis description can be edited later.",
                        "emoji": True,
                    },
                    "multiline": True,
                    "min_length": 10,
                    "max_length": 1200,
                    "dispatch_action_config": {
                        "trigger_actions_on": ["on_enter_pressed"]
                    },
                },
            },
            {
                "type": "input",
                "block_id": "incident_category",
                "label": {"type": "plain_text", "text": "Issue category", "emoji": True},
                "optional": False,
                "dispatch_action": False,
                "element": {
                    "type": "static_select",
                    "action_id": "incident_category",
                    "placeholder": {
                        "type": "plain_text",
                        "text": "Select affected issue category",
                        "emoji": True,
                    },
                    "option_groups": [
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Tech Platform",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Back Office",
                                        "emoji": True,
                                    },
                                    "value": "43b582f1-d445-4fc9-92eb-e742ee14fd4a",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "CDN",
                                        "emoji": True,
                                    },
                                    "value": "962bf559-5fc2-4cd4-8834-6c3c1e767085",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Data Factory",
                                        "emoji": True,
                                    },
                                    "value": "63a9d7d7-6df7-4771-8ad1-a53d2a78951f",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Infrastructure",
                                        "emoji": True,
                                    },
                                    "value": "5abb78d3-1214-4a06-81a1-04351a5c9b33",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "QA",
                                        "emoji": True,
                                    },
                                    "value": "79fa2929-d44f-4c4e-862f-e8d50fa9d5b2",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Visitors",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "B2C Advice",
                                        "emoji": True,
                                    },
                                    "value": "18ac3799-4559-4a50-8964-c89f475265fe",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "B2C Retention",
                                        "emoji": True,
                                    },
                                    "value": "a218d573-4cc9-4862-8eb1-800f826a2f9c",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Search",
                                        "emoji": True,
                                    },
                                    "value": "8434e40e-66e3-4255-ba42-652d5af078c0",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Spartacux",
                                        "emoji": True,
                                    },
                                    "value": "421927b9-76a2-43b3-9a44-92e794b5c523",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Buyers",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Customer Invoicing",
                                        "emoji": True,
                                    },
                                    "value": "e1d539ce-e02b-47fa-b16c-39d26e3ddf9c",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Customer Management",
                                        "emoji": True,
                                    },
                                    "value": "2f8a8cad-4950-41b3-8fb3-ec0b6608f324",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Inventory",
                                        "emoji": True,
                                    },
                                    "value": "4ef370d3-73dc-4ea6-829b-4cd155abaa93",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Shopping Cart",
                                        "emoji": True,
                                    },
                                    "value": "d81c2362-f44c-4e2c-a431-6a5e1c53415b",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Sellers",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Ads management",
                                        "emoji": True,
                                    },
                                    "value": "189daae3-1391-4f84-9aaf-72c587df810e",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "HUB",
                                        "emoji": True,
                                    },
                                    "value": "fb50fa59-4d7f-42ea-99e3-412f63fd91bd",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Sellers Animation",
                                        "emoji": True,
                                    },
                                    "value": "4c896455-f6d8-42c8-b6cf-d5be87cdfc3c",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Sellers Management",
                                        "emoji": True,
                                    },
                                    "value": "89a1798d-b03e-4d36-aa26-b9a7e22379a1",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Toolbox",
                                        "emoji": True,
                                    },
                                    "value": "26dbf9e7-4f09-4f12-a77c-bfb7cc426804",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Product Catalog",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Product Catalog",
                                        "emoji": True,
                                    },
                                    "value": "276be9a2-8ac2-4a84-9ba8-39a4244bfc0b",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Product Delivery",
                                        "emoji": True,
                                    },
                                    "value": "6e4ce13e-527d-426a-a753-36c5f207a8bb",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Product Discovery",
                                        "emoji": True,
                                    },
                                    "value": "86ae1442-a15a-4545-98bb-1a3226b98a23",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Product Feed",
                                        "emoji": True,
                                    },
                                    "value": "3d71d900-bf53-47e2-883b-332c641f7541",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Product Offer  Management",
                                        "emoji": True,
                                    },
                                    "value": "6a0e0920-c67e-451d-8189-bb5d01baf797",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Pro",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Pro Animation",
                                        "emoji": True,
                                    },
                                    "value": "36444a95-6194-4873-8805-0aa4f0e3d5b1",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Pro Management",
                                        "emoji": True,
                                    },
                                    "value": "46fea112-5861-467e-b3d9-50911fe6d567",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Pro Services",
                                        "emoji": True,
                                    },
                                    "value": "b7d8176c-c2d6-4165-9cff-e1edec16bec4",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Finance Operations",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Payment",
                                        "emoji": True,
                                    },
                                    "value": "02a18a9c-67ba-45e8-9b50-8662f1c6eae1",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Seller Finance Operations",
                                        "emoji": True,
                                    },
                                    "value": "59067b04-bf1c-4280-b780-19fd81a236bb",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Sellers Billing",
                                        "emoji": True,
                                    },
                                    "value": "6c63ca62-3b5a-4b68-976b-f396881d84df",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Traffic Acquisition",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Affiliation",
                                        "emoji": True,
                                    },
                                    "value": "f28a2d22-fa77-4d33-8b78-30e3a1828c5e",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Free Acquisition",
                                        "emoji": True,
                                    },
                                    "value": "22e1e81d-d995-435c-9138-4bc721f08a02",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Paid Acquisition",
                                        "emoji": True,
                                    },
                                    "value": "44cfd60f-c904-4e46-a3c9-dab977d0321e",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "SEO",
                                        "emoji": True,
                                    },
                                    "value": "f6f05447-4444-4abb-88b6-728493de19fa",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Mobile",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Mobile B2C Apps",
                                        "emoji": True,
                                    },
                                    "value": "0b7df3de-e3b5-4924-bcff-bb0ccc31fa03",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Mobile PRO Apps",
                                        "emoji": True,
                                    },
                                    "value": "a654d411-4962-48d0-a2b0-7162f82d9c73",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Order",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Customer Order",
                                        "emoji": True,
                                    },
                                    "value": "7529a8b4-ec9b-4722-a772-e10ff7a616ad",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Purchase Order",
                                        "emoji": True,
                                    },
                                    "value": "0c6eafc4-030d-406b-857d-457626a03d25",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Seller Order",
                                        "emoji": True,
                                    },
                                    "value": "eef1c9d0-a460-4821-8dc7-6361ba1e016c",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Validate Order",
                                        "emoji": True,
                                    },
                                    "value": "e64edcbf-1908-49fe-82d7-6c2e4d4b54f4",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Security",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Attack",
                                        "emoji": True,
                                    },
                                    "value": "0dff9974-b5bd-4d3d-84a0-b93b8d3a69ce",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Security Misc",
                                        "emoji": True,
                                    },
                                    "value": "89c9ff50-5b0b-4554-a3e3-cbe31258dd19",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "System Compromise",
                                        "emoji": True,
                                    },
                                    "value": "8bfafb93-8790-42f0-898a-d2e30e8ba532",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Care",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Customer Care",
                                        "emoji": True,
                                    },
                                    "value": "f460e4ee-ac13-4b94-b9e4-fbb0aa87558c",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Internal Messaging",
                                        "emoji": True,
                                    },
                                    "value": "de920235-2c36-45fc-b366-02e9079603b2",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Sysadmin",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Applications",
                                        "emoji": True,
                                    },
                                    "value": "0f2c7343-d526-4aee-8b54-ef1de7c1fd61",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Infra - System",
                                        "emoji": True,
                                    },
                                    "value": "3e0bedcb-725b-4c02-876f-c7665d5bb15d",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Network",
                                        "emoji": True,
                                    },
                                    "value": "c419c3a0-9490-416e-8b4c-4e39c55b41a8",
                                },
                            ],
                        },
                        {
                            "label": {
                                "type": "plain_text",
                                "text": "Other",
                                "emoji": True,
                            },
                            "options": [
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Legal",
                                        "emoji": True,
                                    },
                                    "value": "5900badb-1d82-4aa6-b6e2-b76f29a83d5f",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Manomano",
                                        "emoji": True,
                                    },
                                    "value": "432e2ea4-4908-484b-8fa8-2bd943cff41f",
                                },
                                {
                                    "text": {
                                        "type": "plain_text",
                                        "text": "Other",
                                        "emoji": True,
                                    },
                                    "value": "390a993a-d273-4db8-b7d6-190ab294961a",
                                },
                            ],
                        },
                    ],
                },
            },
            {
                "type": "input",
                "block_id": "severity",
                "label": {"type": "plain_text", "text": "Severity", "emoji": True},
                "optional": False,
                "dispatch_action": False,
                "element": {
                    "type": "static_select",
                    "action_id": "severity",
                    "placeholder": {
                        "type": "plain_text",
                        "text": "Select a severity",
                        "emoji": True,
                    },
                    "initial_option": {
                        "text": {
                            "type": "plain_text",
                            "text": "🌦️  SEV4 - Minor issue not affecting customers.",
                            "emoji": True,
                        },
                        "value": "b814c9d2-48a8-4ac4-9c71-ff844e1b77f1",
                    },
                    "options": [
                        {
                            "text": {
                                "type": "plain_text",
                                "text": "🌪️  SEV1 - Critical issue that warrants liaison with e-team.",
                                "emoji": True,
                            },
                            "value": "3d8a4e0f-cc79-4063-9431-8aa785a56ca0",
                        },
                        {
                            "text": {
                                "type": "plain_text",
                                "text": "⛈️  SEV2 - Major issue impacting customers.",
                                "emoji": True,
                            },
                            "value": "8ab02fa9-f950-49cb-982e-b8d653e38ffc",
                        },
                        {
                            "text": {
                                "type": "plain_text",
                                "text": "🌩️  SEV3 - Minor issue affecting customers.",
                                "emoji": True,
                            },
                            "value": "e12c3836-ea9b-44bf-a3cf-5a715d62395b",
                        },
                        {
                            "text": {
                                "type": "plain_text",
                                "text": "🌦️  SEV4 - Minor issue not affecting customers.",
                                "emoji": True,
                            },
                            "value": "b814c9d2-48a8-4ac4-9c71-ff844e1b77f1",
                        },
                        {
                            "text": {
                                "type": "plain_text",
                                "text": "🐞  SEV5 - Cosmetic issue not affecting customers.",
                                "emoji": True,
                            },
                            "value": "e57c2084-d7e0-4dea-9f6b-a4e8bc5a55d4",
                        },
                        {
                            "text": {
                                "type": "plain_text",
                                "text": "🤡  GAMEDAY - Gamedays experiments",
                                "emoji": True,
                            },
                            "value": "701d498c-067f-43cd-b1ca-0977d983f27e",
                        },
                    ],
                },
            },
            {
                "type": "section",
                "block_id": "fTkH",
                "text": {
                    "type": "mrkdwn",
                    "text": "_<https://example.atlassian.net/wiki/spaces/INCIDENT/pages/123456>_",
                    "verbatim": False,
                },
            },
            {
                "type": "input",
                "block_id": "environment",
                "label": {"type": "plain_text", "text": "Environment", "emoji": True},
                "optional": False,
                "dispatch_action": False,
                "element": {
                    "type": "static_select",
                    "action_id": "environment",
                    "placeholder": {
                        "type": "plain_text",
                        "text": "Select an environment",
                        "emoji": True,
                    },
                    "initial_option": {
                        "text": {
                            "type": "plain_text",
                            "text": "PRD - Production environment",
                            "emoji": True,
                        },
                        "value": "e5c46eb0-5620-4a30-81c6-4a647c189ff2",
                    },
                    "options": [
                        {
                            "text": {
                                "type": "plain_text",
                                "text": "INT - Integration environment",
                                "emoji": True,
                            },
                            "value": "1b960430-995b-47e1-beab-23dbe3dbccbf",
                        },
                        {
                            "text": {
                                "type": "plain_text",
                                "text": "STG - Staging environment",
                                "emoji": True,
                            },
                            "value": "41e06f3b-4b42-4617-af9a-fa91e9af13c9",
                        },
                        {
                            "text": {
                                "type": "plain_text",
                                "text": "PRD - Production environment",
                                "emoji": True,
                            },
                            "value": "e5c46eb0-5620-4a30-81c6-4a647c189ff2",
                        },
                    ],
                },
            },
        ],
        "private_metadata": "",
        "callback_id": "incident_close",
        "state": {
            "values": {
                "title": {
                    "title": {"type": "plain_text_input", "value": "Valid incident"}
                },
                "description": {
                    "description": {
                        "type": "plain_text_input",
                        "value": "This is a valid description.",
                    }
                },
                "incident_category": {
                    "incident_category": {
                        "type": "static_select",
                        "selected_option": {
                            "text": {
                                "type": "plain_text",
                                "text": "Back Office",
                                "emoji": True,
                            },
                            "value": "43b582f1-d445-4fc9-92eb-e742ee14fd4a",
                        },
                    }
                },
                "severity": {
                    "severity": {
                        "type": "static_select",
                        "selected_option": {
                            "text": {
                                "type": "plain_text",
                                "text": "🌦️  SEV4 - Minor issue not affecting customers.",
                                "emoji": True,
                            },
                            "value": "b814c9d2-48a8-4ac4-9c71-ff844e1b77f1",
                        },
                    }
                },
                "environment": {
                    "environment": {
                        "type": "static_select",
                        "selected_option": {
                            "text": {
                                "type": "plain_text",
                                "text": "PRD - Production environment",
                                "emoji": True,
                            },
                            "value": "e5c46eb0-5620-4a30-81c6-4a647c189ff2",
                        },
                    }
                },
            }
        },
        "hash": "1660123198.SvjyXz97",
        "title": {"type": "plain_text", "text": "Close an incident", "emoji": True},
        "clear_on_close": False,
        "notify_on_close": False,
        "close": None,
        "submit": {"type": "plain_text", "text": "Close an incident", "emoji": True},
        "previous_view_id": None,
        "root_view_id": "V03T2SMMWBV",
        "app_id": "A020BG97EG5",
        "external_id": "",
        "app_installed_team_id": "T01FJ0NNFQD",
        "bot_id": "B020T8C8U9F",
    },
    "response_urls": [],
    "is_enterprise_install": False,
    "enterprise": None,
}


invalid_title = deepcopy(valid_submission)
invalid_title["view"]["state"]["values"]["title"]["title"]["value"] = "short"  # type: ignore

invalid_incident_category = deepcopy(valid_submission)
invalid_incident_category["view"]["state"]["values"]["incident_category"]["incident_category"][
    "selected_option"
]["value"] = "notauuid-d445-4fc9-92eb-e742ee14fd4a"  # type: ignore



"""Test complete workflow transitions according to the diagram.

Workflow transitions:

P1/P2:
- OPEN → [INVESTIGATING, CLOSED (avec reason form)]
- INVESTIGATING → [MITIGATING, CLOSED (avec reason form)]
- MITIGATING → [MITIGATED]
- MITIGATED → [POST_MORTEM]
- POST_MORTEM → [CLOSED]

P3/P4/P5:
- OPEN → [INVESTIGATING, CLOSED (avec reason form)]
- INVESTIGATING → [MITIGATING, CLOSED (avec reason form)]
- MITIGATING → [MITIGATED]
- MITIGATED → [CLOSED]
"""
from __future__ import annotations

import pytest
from django.test import TestCase

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.factories import IncidentFactory
from firefighter.incidents.forms.update_status import UpdateStatusForm
from firefighter.incidents.models import Environment, Priority


@pytest.mark.django_db
class TestCompleteWorkflowTransitions(TestCase):
    """Test that all workflow transitions are implemented correctly."""

    def test_p1_p2_complete_workflow_transitions(self):
        """Test all P1/P2 workflow transitions according to the diagram."""
        # Get P1/P2 priority
        p1_priority = Priority.objects.filter(
            value__in=[1, 2], needs_postmortem=True
        ).first()

        if not p1_priority:
            pytest.skip("No P1/P2 priority found")

        prd_env = Environment.objects.filter(value="PRD").first()
        if not prd_env:
            pytest.skip("No PRD environment found")

        # Test all transitions
        transitions = [
            # Test transitions from each status
            (IncidentStatus.OPEN, [IncidentStatus.INVESTIGATING, IncidentStatus.CLOSED]),
            (IncidentStatus.INVESTIGATING, [IncidentStatus.MITIGATING, IncidentStatus.CLOSED]),
            (IncidentStatus.MITIGATING, [IncidentStatus.MITIGATED]),
            (IncidentStatus.MITIGATED, [IncidentStatus.POST_MORTEM, IncidentStatus.INVESTIGATING, IncidentStatus.MITIGATING]),
            (IncidentStatus.POST_MORTEM, [IncidentStatus.CLOSED]),
        ]

        for current_status, expected_statuses in transitions:
            incident = IncidentFactory.create(
                priority=p1_priority,
                environment=prd_env,
                _status=current_status,
            )

            form = UpdateStatusForm(incident=incident)
            status_choices = dict(form.fields["status"].choices)
            # Convert string keys back to enum values for comparison
            available_statuses = [IncidentStatus(int(k)) for k in status_choices]

            # Remove current status from available (can't transition to same status)
            if current_status in available_statuses:
                available_statuses.remove(current_status)

            for expected_status in expected_statuses:
                assert expected_status in available_statuses, (
                    f"P1/P2: From {current_status.label}, should be able to go to {expected_status.label}. "
                    f"Available: {[s.label for s in available_statuses]}"
                )

            # Verify no unexpected statuses are available
            unexpected_statuses = set(available_statuses) - set(expected_statuses)
            assert not unexpected_statuses, (
                f"P1/P2: From {current_status.label}, unexpected statuses available: "
                f"{[s.label for s in unexpected_statuses]}"
            )

    def test_p3_plus_complete_workflow_transitions(self):
        """Test all P3+ workflow transitions according to the diagram."""
        # Get P3+ priority
        p3_priority = Priority.objects.filter(
            value__gte=3, needs_postmortem=False
        ).first()

        if not p3_priority:
            pytest.skip("No P3+ priority found")

        # Test all transitions
        transitions = [
            # Test transitions from each status
            (IncidentStatus.OPEN, [IncidentStatus.INVESTIGATING, IncidentStatus.CLOSED]),
            (IncidentStatus.INVESTIGATING, [IncidentStatus.MITIGATING, IncidentStatus.CLOSED]),
            (IncidentStatus.MITIGATING, [IncidentStatus.MITIGATED]),
            (IncidentStatus.MITIGATED, [IncidentStatus.CLOSED, IncidentStatus.INVESTIGATING, IncidentStatus.MITIGATING]),
        ]

        for current_status, expected_statuses in transitions:
            incident = IncidentFactory.create(
                priority=p3_priority,
                _status=current_status,
            )

            form = UpdateStatusForm(incident=incident)
            status_choices = dict(form.fields["status"].choices)
            # Convert string keys back to enum values for comparison
            available_statuses = [IncidentStatus(int(k)) for k in status_choices]

            # Remove current status from available (can't transition to same status)
            if current_status in available_statuses:
                available_statuses.remove(current_status)

            for expected_status in expected_statuses:
                assert expected_status in available_statuses, (
                    f"P3+: From {current_status.label}, should be able to go to {expected_status.label}. "
                    f"Available: {[s.label for s in available_statuses]}"
                )

            # Verify no unexpected statuses are available (especially POST_MORTEM)
            unexpected_statuses = set(available_statuses) - set(expected_statuses)
            assert not unexpected_statuses, (
                f"P3+: From {current_status.label}, unexpected statuses available: "
                f"{[s.label for s in unexpected_statuses]}"
            )

            # Specifically verify POST_MORTEM is never available for P3+
            assert IncidentStatus.POST_MORTEM not in available_statuses, (
                f"P3+: POST_MORTEM should never be available, but found from {current_status.label}"
            )

    def test_closure_reason_requirements(self):
        """Test that closure reason is required only from OPEN and INVESTIGATING."""
        priorities = [
            Priority.objects.filter(value__in=[1, 2]).first(),  # P1/P2
            Priority.objects.filter(value__gte=3).first(),       # P3+
        ]

        for priority in priorities:
            if not priority:
                continue

            # Should require reason from OPEN and INVESTIGATING
            for status in [IncidentStatus.OPEN, IncidentStatus.INVESTIGATING]:
                incident = IncidentFactory.create(
                    priority=priority,
                    _status=status,
                )

                assert UpdateStatusForm.requires_closure_reason(
                    incident, IncidentStatus.CLOSED
                ), f"Should require closure reason from {status.label} for {priority.name}"

            # Should NOT require reason from other statuses
            for status in [IncidentStatus.MITIGATING, IncidentStatus.MITIGATED, IncidentStatus.POST_MORTEM]:
                incident = IncidentFactory.create(
                    priority=priority,
                    _status=status,
                )

                assert not UpdateStatusForm.requires_closure_reason(
                    incident, IncidentStatus.CLOSED
                ), f"Should NOT require closure reason from {status.label} for {priority.name}"



"""Django settings for server project.

For more information on this file, see
https://docs.djangoproject.com/en/4.2/topics/settings/

For the full list of settings and their config, see
https://docs.djangoproject.com/en/4.2/ref/settings/
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

from decouple import Csv

from firefighter.firefighter.settings.settings_utils import _ENV, BASE_DIR, config
from firefighter.firefighter.sso import link_auth_user

if TYPE_CHECKING:
    from collections.abc import Callable

SECRET_KEY: str = config("SECRET_KEY")
ENV = _ENV

# Application definition:
INSTALLED_APPS = [
    # Your apps go here:
    "firefighter.firefighter",
    "firefighter.incidents",
    "firefighter.slack",
    # Default django apps
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.humanize",
    "django.contrib.postgres",
    "django.forms",
    # django-admin
    "django.contrib.admin",
    "django.contrib.admindocs",
    # API
    "firefighter.api",
    "rest_framework",
    "rest_framework.authtoken",
    "drf_spectacular",
    "drf_standardized_errors",
    # Other deps
    "widget_tweaks",
    "django_components",
    "django_htmx",
    "django_filters",
    "taggit",
    "django_tables2",
    "import_export",
    # SSO Auth
    "oauth2_authcodeflow",
    # Celery integration
    "django_celery_beat",
    # Menu generation from class
    "simple_menu",
]

MIDDLEWARE: tuple[str, ...] = (
    "django.middleware.security.SecurityMiddleware",
    # Whitenoise (serves assets) must come first, but after django.middleware.security.SecurityMiddleware
    "whitenoise.middleware.WhiteNoiseMiddleware",
    # Django:
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    # XXX We could make a single Middleware to replace AuthenticationMiddleware and LoginRequiredMiddleware
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "oauth2_authcodeflow.middleware.LoginRequiredMiddleware",
    "django_htmx.middleware.HtmxMiddleware",
    "firefighter.firefighter.middleware.HeaderUser",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
)

ROOT_URLCONF = "firefighter.firefighter.urls"

WSGI_APPLICATION = "firefighter.firefighter.wsgi.application"


# Database
# https://docs.djangoproject.com/en/4.2/ref/settings/#databases

db_schema = config("POSTGRES_SCHEMA", default="")
db_options = {"options": "-c statement_timeout=30000"}
if db_schema:
    db_options["options"] += f" -c search_path={db_schema}"

DATABASES: dict[str, dict[str, Any]] = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": config("POSTGRES_DB"),
        "USER": config("POSTGRES_USER"),
        "PASSWORD": config("POSTGRES_PASSWORD"),
        "HOST": config("POSTGRES_HOST"),
        "PORT": config("POSTGRES_PORT"),
        "OPTIONS": db_options,
    }
}

# Internationalization
# https://docs.djangoproject.com/en/4.2/topics/i18n/

LANGUAGE_CODE = "en"

USE_I18N = False
USE_TZ = True
TIME_ZONE: str = config("TIME_ZONE", default="UTC")

# Localization
FORMAT_MODULE_PATH = ["firefighter.formats"]

# Default date filters
# Set in firefighter.formats as well
SHORT_DATE_FORMAT = "Y/m/d"
SHORT_DATETIME_FORMAT = "Y/m/d H:i"


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.2/howto/static-files/


STATIC_URL = "/static/"
STATIC_ROOT = BASE_DIR / ".static"
STORAGES = {
    "staticfiles": {
        "BACKEND": "whitenoise.storage.CompressedManifestStaticFilesStorage",
    },
}
STATICFILES_DIRS = [
    BASE_DIR / "components",
]
# Templates
# https://docs.djangoproject.com/en/4.2/ref/templates/api


TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [
            Path(BASE_DIR.joinpath("templates")),
            Path(BASE_DIR.joinpath("components")),
        ],
        "APP_DIRS": False,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "firefighter.firefighter.utils.get_global_context",
            ],
            "loaders": [
                (
                    "django.template.loaders.cached.Loader",
                    [
                        "django.template.loaders.filesystem.Loader",
                        "django.template.loaders.app_directories.Loader",
                        "django_components.template_loader.Loader",
                    ],
                )
            ],
            "builtins": [
                "django_components.templatetags.component_tags",
            ],
        },
    },
]
FORM_RENDERER = "django.forms.renderers.TemplatesSetting"


# Django authentication system
# https://docs.djangoproject.com/en/4.2/topics/auth/

AUTHENTICATION_BACKENDS = (
    "oauth2_authcodeflow.auth.AuthenticationBackend",
    "django.contrib.auth.backends.ModelBackend",
)


# Security
# https://docs.djangoproject.com/en/4.2/topics/security/

# Password validation
# https://docs.djangoproject.com/en/4.2/ref/settings/#auth-password-validators

_PASS = "django.contrib.auth.password_validation"  # noqa: S105 # nosec
AUTH_PASSWORD_VALIDATORS = [
    {"NAME": f"{_PASS}.UserAttributeSimilarityValidator"},
    {"NAME": f"{_PASS}.MinimumLengthValidator"},
    {"NAME": f"{_PASS}.CommonPasswordValidator"},
    {"NAME": f"{_PASS}.NumericPasswordValidator"},
]

ALLOWED_HOSTS = ["*"]
CSRF_TRUSTED_ORIGINS: list[str] = config(
    "CSRF_TRUSTED_ORIGINS", default="https://*", cast=Csv()
)


SESSION_COOKIE_HTTPONLY = True
CSRF_COOKIE_HTTPONLY = True
SECURE_CONTENT_TYPE_NOSNIFF = True
X_FRAME_OPTIONS = "DENY"

# Password hashing
# https://docs.djangoproject.com/en/4.2/topics/auth/passwords/#auth-password-storage
# Argon2 + defaults
PASSWORD_HASHERS = [
    "django.contrib.auth.hashers.Argon2PasswordHasher",
    "django.contrib.auth.hashers.PBKDF2PasswordHasher",
    "django.contrib.auth.hashers.PBKDF2SHA1PasswordHasher",
    "django.contrib.auth.hashers.BCryptSHA256PasswordHasher",
    "django.contrib.auth.hashers.ScryptPasswordHasher",
]

# Default primary key field type
# https://docs.djangoproject.com/en/4.2/ref/settings/#default-auto-field
DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

# Can be removed when we upgrade to Django 5.0, which makes this the default
FORM_RENDERER = "django.forms.renderers.DjangoDivFormRenderer"

# Debug Flags
DEBUG_TOOLBAR = config("DEBUG_TOOLBAR", default=False, cast=bool)

USE_X_FORWARDED_HOST = True


#  Auth and SSO config
def get_django_username(x: dict[str, str]) -> str | None:
    username = x.get("preferred_username")
    if username is not None and "@" in username:
        return username.split("@")[0]
    return username


AUTH_USER_MODEL = "incidents.User"
OIDC_TIMEOUT: int = config("OIDC_TIMEOUT", cast=int, default=15)
OIDC_OP_DISCOVERY_DOCUMENT_URL = config("OIDC_OP_DISCOVERY_DOCUMENT_URL")
OIDC_RP_CLIENT_ID = config("OIDC_RP_CLIENT_ID")
OIDC_RP_CLIENT_SECRET = config("OIDC_RP_CLIENT_SECRET")
OIDC_MIDDLEWARE_NO_AUTH_URL_PATTERNS = [
    "^/admin/",
    "^/__reload__/events/",
    "^/api/",
    "^/err/",
    "^/robots.txt$",
]

# noinspection PyPep8
OIDC_DJANGO_USERNAME_FUNC = get_django_username
OIDC_CREATE_USER = True
OIDC_EXTEND_USER = link_auth_user
OIDC_UNUSABLE_PASSWORD = config("OIDC_UNUSABLE_PASSWORD", default=False, cast=bool)
OIDC_RP_USE_PKCE = config("OIDC_RP_USE_PKCE", default=False, cast=bool)
# Default login redirect
LOGIN_URL = "/admin/login/"

# Plausible analytics configuration

PLAUSIBLE_SCRIPT_URL: str = config("PLAUSIBLE_SCRIPT_URL", "")
PLAUSIBLE_DOMAIN: str = config("PLAUSIBLE_DOMAIN", "")


# Components
COMPONENTS = {
    "autodiscover": False,
    "context_behavior": "django",  # Like before django-components 0.67
    "libraries": [
        "firefighter.components.avatar.avatar",
        "firefighter.components.card.card",
        "firefighter.components.export_button.export_button",
        "firefighter.components.form.form",
        "firefighter.components.form_field.form_field",
        "firefighter.components.modal.modal",
        "firefighter.components.messages.messages",
    ],
}


# FF specific options
FF_ROLE_REMINDER_MIN_DAYS_INTERVAL = config(
    "FF_ROLE_REMINDER_MIN_DAYS_INTERVAL", default=90, cast=int
)
"Number of days between role explanation/reminders, for each role. -1 disable the messages, and 0 will send the message everytime."

FF_HTTP_CLIENT_ADDITIONAL_HEADERS: dict[str, Any] | None = None
"Additional headers to send with every HTTP request made using our HttpClient. Useful for global auth, or adding a specific User-Agent."

FF_USER_ID_HEADER: str = "FF-User-Id"
"Header name to add to every HTTP request made using our HttpClient. Useful for logging."

FF_OVERRIDE_MENUS_CREATION: Callable[[], None] | None = None
"Override the default menus creation. Useful for custom menus."

FF_DEBUG_ERROR_PAGES: bool = config("FF_DEBUG_ERROR_PAGES", default=False, cast=bool)
"Add routes to display error pages. Useful for debugging."

FF_SKIP_SECRET_KEY_CHECK: bool = config(
    "FF_SKIP_SECRET_KEY_CHECK", default=False, cast=bool
)
"Skip the SECRET_KEY check. Make sure to set a strong SECRET_KEY in production."

FF_EXPOSE_API_DOCS: bool = config("FF_EXPOSE_API_DOCS", default=False, cast=bool)
"Expose the API documentation. Useful for debugging. Can be a security issue."



from __future__ import annotations

from django.conf import settings

from firefighter.pagerduty.signals.get_invites_from_pagerduty import (
    get_invites_from_pagerduty,
)

if settings.ENABLE_SLACK:
    from firefighter.pagerduty.signals.incident_channel_done_oncall import (
        incident_channel_done_oncall,
    )



from __future__ import annotations

import logging
from functools import cache
from textwrap import TextWrapper
from typing import TYPE_CHECKING, Any, TypeVar

from django.conf import settings
from django.utils.timezone import localtime
from slack_sdk.models.blocks.basic_components import MarkdownTextObject
from slack_sdk.models.blocks.blocks import ContextBlock, DividerBlock, SectionBlock

from firefighter.slack.models.conversation import Conversation

if TYPE_CHECKING:
    from datetime import datetime

    from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)

COMMAND: str = settings.SLACK_INCIDENT_COMMAND
INCIDENT_DOC_URL: str | None = settings.SLACK_INCIDENT_HELP_GUIDE_URL


def shorten_long(text: str, width: int, **kwargs: Any) -> str:
    """Shorten text while keeping newlines and most formatting."""
    kwargs.setdefault("placeholder", " [...]")
    kwargs.setdefault("tabsize", 4)
    kwargs.setdefault("max_lines", 1)
    kwargs.setdefault("drop_whitespace", False)
    kwargs.setdefault("replace_whitespace", False)
    kwargs.setdefault("break_long_words", True)
    w = TextWrapper(
        width=width,
        **kwargs,
    )
    return w.fill(text)


T = TypeVar("T")


def md_quote_filter[T](val: str | T) -> str | T:
    """Add > on newlines for MD quotes."""
    if isinstance(val, str):
        return val.replace("\n", "\n> ")
    return val


@cache
def date_time(date: datetime | None) -> str:
    """Common format for datetime.

    Args:
        date (datetime | None): your datetime

    Returns:
        str: datetime in format `YYYY-MM-DD HH:MM`
    """
    return localtime(date).strftime("%Y-%m-%d %H:%M")


def user_slack_handle_or_name(user: User | None) -> str:
    """Returns the Slack handle of the user in Slack MD format (`<@SLACK_ID>`) or the user full name."""
    if user is None:
        return "∅"

    if hasattr(user, "slack_user") and user.slack_user:
        return f"<@{user.slack_user.slack_id}>"
    return user.full_name


@cache
def slack_block_footer() -> ContextBlock:
    support_channel = Conversation.objects.get_or_none(tag="dev_firefighter")
    support_text = (
        f" Support and feedback in <#{support_channel.channel_id}>"
        if support_channel
        else ""
    )
    return ContextBlock(
        elements=[
            MarkdownTextObject(
                text=f"{settings.SLACK_APP_EMOJI}  {settings.APP_DISPLAY_NAME} {settings.FF_VERSION}. {support_text}"
            )
        ]
    )


@cache
def slack_block_separator() -> DividerBlock:
    return DividerBlock()


@cache
def slack_block_help_commands() -> SectionBlock:
    return SectionBlock(
        text=f"- `{COMMAND} open`: open an incident to start investigation\n- `{COMMAND} update`: update incident roles or statuses\n- `{COMMAND} close`: close an incident and archive the channel\n- `{COMMAND} status`: get a recap of the incident from this channel\n- `{COMMAND} oncall`: select an on-call you want to call\n- `{COMMAND} postmortem`: create the postmortem if needed\n- `{COMMAND} sos`: ask for SRE help"
    )


@cache
def slack_block_help_description() -> SectionBlock:
    if INCIDENT_DOC_URL is None:
        return SectionBlock(
            text=f"{settings.APP_DISPLAY_NAME} is our tool for incident management."
        )
    return SectionBlock(
        text=f"{settings.APP_DISPLAY_NAME} is our tool for incident management, more about incidents is visible <{INCIDENT_DOC_URL}|here>."
    )


@cache
def slack_block_help_tip() -> SectionBlock:
    return SectionBlock(
        text="A good incident response process involves great communication with internal and external stakeholders!"
    )


def slack_block_quote(text: str, length: int = 2995) -> SectionBlock:
    return SectionBlock(text=f"> {shorten_long(md_quote_filter(text), length)}")



"""Tests for raid.service module."""

from __future__ import annotations

from unittest.mock import Mock, patch

import pytest

from firefighter.jira_app.client import JiraAPIError, JiraUserNotFoundError
from firefighter.raid.service import (
    CustomerIssueData,
    check_issue_id,
    create_issue_customer,
    create_issue_documentation_request,
    create_issue_feature_request,
    create_issue_internal,
    create_issue_seller,
    get_jira_user_from_user,
)


class TestCustomerIssueData:
    """Test the CustomerIssueData dataclass."""

    def test_customer_issue_data_creation(self):
        """Test creating CustomerIssueData with all fields."""
        data = CustomerIssueData(
            priority=1,
            labels=["urgent", "customer"],
            platform="web",
            business_impact="high",
            team_to_be_routed="backend-team",
            area="payments",
            zendesk_ticket_id="12345",
            incident_category="payment-issue",
        )

        assert data.priority == 1
        assert data.labels == ["urgent", "customer"]
        assert data.platform == "web"
        assert data.business_impact == "high"
        assert data.team_to_be_routed == "backend-team"
        assert data.area == "payments"
        assert data.zendesk_ticket_id == "12345"
        assert data.incident_category == "payment-issue"

    def test_customer_issue_data_defaults(self):
        """Test CustomerIssueData with default values."""
        data = CustomerIssueData(
            priority=None,
            labels=None,
            platform="mobile",
            business_impact=None,
            team_to_be_routed=None,
            area=None,
            zendesk_ticket_id=None,
        )

        assert data.priority is None
        assert data.labels is None
        assert data.platform == "mobile"
        assert data.business_impact is None
        assert data.team_to_be_routed is None
        assert data.area is None
        assert data.zendesk_ticket_id is None
        assert data.incident_category is None  # Default value


@pytest.mark.django_db
class TestGetJiraUserFromUser:
    """Test get_jira_user_from_user function."""

    @patch("firefighter.raid.service.jira_client")
    def test_get_jira_user_success(self, mock_jira_client, admin_user):
        """Test successful get_jira_user_from_user."""
        mock_jira_user = Mock()
        mock_jira_user.id = "test_jira_id"
        mock_jira_client.get_jira_user_from_user.return_value = mock_jira_user

        result = get_jira_user_from_user(admin_user)

        mock_jira_client.get_jira_user_from_user.assert_called_once_with(admin_user)
        assert result == mock_jira_user

    @patch("firefighter.raid.service.jira_client")
    def test_get_jira_user_fallback_to_default(self, mock_jira_client, admin_user):
        """Test fallback to default user when jira_client fails."""

        # Make the first call fail
        mock_jira_client.get_jira_user_from_user.side_effect = JiraAPIError("API error")

        # Mock the fallback call to succeed
        mock_fallback_user = Mock()
        mock_fallback_user.id = "fallback_id"
        mock_jira_client.get_jira_user_from_jira_id.return_value = mock_fallback_user

        result = get_jira_user_from_user(admin_user)

        # Should call the fallback method
        mock_jira_client.get_jira_user_from_jira_id.assert_called_once()
        assert result == mock_fallback_user

    @patch("firefighter.raid.service.jira_client")
    @patch("firefighter.raid.service.JiraUser")
    def test_get_jira_user_fallback_to_db(self, mock_jira_user_model, mock_jira_client, admin_user):
        """Test fallback to database when both API calls fail."""

        # Make both API calls fail
        mock_jira_client.get_jira_user_from_user.side_effect = JiraAPIError("API error")
        mock_jira_client.get_jira_user_from_jira_id.side_effect = JiraUserNotFoundError("User not found")

        # Mock database fallback
        mock_db_user = Mock()
        mock_db_user.id = "db_user_id"
        mock_jira_user_model.objects.get.return_value = mock_db_user

        result = get_jira_user_from_user(admin_user)

        # Should try both API calls then fallback to DB
        mock_jira_client.get_jira_user_from_user.assert_called_once_with(admin_user)
        mock_jira_client.get_jira_user_from_jira_id.assert_called_once()
        mock_jira_user_model.objects.get.assert_called_once()
        assert result == mock_db_user


class TestCheckIssueId:
    """Test check_issue_id function."""

    def test_check_issue_id_with_valid_issue(self):
        """Test check_issue_id with valid issue object."""
        mock_issue = {"id": "TICKET-123"}

        result = check_issue_id(mock_issue, "Test Title", "test_reporter")
        assert result == "TICKET-123"

    def test_check_issue_id_with_none_issue(self):
        """Test check_issue_id with None issue should raise AttributeError."""
        # The actual code doesn't check for None, so it raises AttributeError
        with pytest.raises(AttributeError):
            check_issue_id(None, "Test Title", "test_reporter")

    def test_check_issue_id_with_issue_without_id(self):
        """Test check_issue_id with issue object without id."""
        mock_issue = {"id": None}

        with pytest.raises(JiraAPIError):
            check_issue_id(mock_issue, "Test Title", "test_reporter")


@pytest.mark.django_db
class TestCreateIssueFunctions:
    """Test the create_issue_* functions."""

    @patch("firefighter.raid.service.jira_client")
    def test_create_issue_customer(self, mock_jira_client):
        """Test create_issue_customer function."""
        # Setup mock
        mock_issue = Mock()
        mock_issue.id = "CUST-123"
        mock_jira_client.create_issue.return_value = mock_issue

        # Create test data
        issue_data = CustomerIssueData(
            priority=1,
            labels=["customer"],
            platform="web",
            business_impact="high",
            team_to_be_routed="support-team",
            area="billing",
            zendesk_ticket_id="ZD-456",
            incident_category="billing-issue",
        )

        # Call function
        result = create_issue_customer(
            title="Customer Issue",
            description="Customer is experiencing billing problems",
            reporter="reporter_id",
            issue_data=issue_data,
        )

        # Verify jira_client.create_issue was called with correct parameters
        mock_jira_client.create_issue.assert_called_once_with(
            issuetype="Incident",
            summary="Customer Issue",
            description="Customer is experiencing billing problems",
            assignee=None,
            reporter="reporter_id",
            priority=1,
            labels=["customer"],
            platform="web",
            business_impact="high",
            suggested_team_routing="support-team",
            zendesk_ticket_id="ZD-456",
            incident_category="billing-issue",
        )

        assert result == mock_issue

    @patch("firefighter.raid.service.jira_client")
    def test_create_issue_feature_request(self, mock_jira_client):
        """Test create_issue_feature_request function."""
        mock_issue = {"id": "FEAT-123"}
        mock_jira_client.create_issue.return_value = mock_issue

        result = create_issue_feature_request(
            title="New Feature",
            description="Feature description",
            reporter="reporter_id",
            priority=2,
            labels=["enhancement"],
            platform="mobile",
        )

        mock_jira_client.create_issue.assert_called_once()
        call_args = mock_jira_client.create_issue.call_args
        assert call_args[1]["issuetype"] == "Feature Request"
        assert call_args[1]["summary"] == "New Feature"
        assert call_args[1]["description"] == "Feature description"
        assert call_args[1]["reporter"] == "reporter_id"
        assert call_args[1]["priority"] == 2

        assert result == mock_issue

    @patch("firefighter.raid.service.jira_client")
    def test_create_issue_feature_request_with_none_labels(self, mock_jira_client):
        """Test create_issue_feature_request with None labels (covers line 75)."""
        mock_issue = {"id": "FEAT-124"}
        mock_jira_client.create_issue.return_value = mock_issue

        result = create_issue_feature_request(
            title="Feature with None labels",
            description="Feature description",
            reporter="reporter_id",
            priority=1,
            labels=None,  # This will trigger line 75: labels = [""]
            platform="web",
        )

        mock_jira_client.create_issue.assert_called_once()
        call_args = mock_jira_client.create_issue.call_args
        # Should have default labels + feature-request
        expected_labels = ["", "feature-request"]
        assert call_args[1]["labels"] == expected_labels
        assert result == mock_issue

    @patch("firefighter.raid.service.jira_client")
    def test_create_issue_feature_request_with_existing_label(self, mock_jira_client):
        """Test create_issue_feature_request when label already exists (covers branch 76->78)."""
        mock_issue = {"id": "FEAT-125"}
        mock_jira_client.create_issue.return_value = mock_issue

        result = create_issue_feature_request(
            title="Feature with existing label",
            description="Feature description",
            reporter="reporter_id",
            priority=1,
            labels=["custom", "feature-request"],  # Already has feature-request
            platform="web",
        )

        mock_jira_client.create_issue.assert_called_once()
        call_args = mock_jira_client.create_issue.call_args
        # Should not duplicate the feature-request label
        expected_labels = ["custom", "feature-request"]
        assert call_args[1]["labels"] == expected_labels
        assert result == mock_issue

    @patch("firefighter.raid.service.jira_client")
    def test_create_issue_documentation_request(self, mock_jira_client):
        """Test create_issue_documentation_request function."""
        mock_issue = {"id": "DOC-123"}
        mock_jira_client.create_issue.return_value = mock_issue

        result = create_issue_documentation_request(
            title="Update Documentation",
            description="Documentation needs updating",
            reporter="reporter_id",
            priority=3,
            labels=["docs"],
            platform="web",
        )

        mock_jira_client.create_issue.assert_called_once()
        call_args = mock_jira_client.create_issue.call_args
        assert call_args[1]["issuetype"] == "Documentation/Process Request"
        assert call_args[1]["summary"] == "Update Documentation"

        assert result == mock_issue

    @patch("firefighter.raid.service.jira_client")
    def test_create_issue_documentation_request_with_none_labels(self, mock_jira_client):
        """Test create_issue_documentation_request with None labels (covers line 111)."""
        mock_issue = {"id": "DOC-124"}
        mock_jira_client.create_issue.return_value = mock_issue

        result = create_issue_documentation_request(
            title="Doc with None labels",
            description="Documentation description",
            reporter="reporter_id",
            priority=2,
            labels=None,  # This will trigger line 111: labels = [""]
            platform="mobile",
        )

        mock_jira_client.create_issue.assert_called_once()
        call_args = mock_jira_client.create_issue.call_args
        # Should have default labels + documentation-request
        expected_labels = ["", "documentation-request"]
        assert call_args[1]["labels"] == expected_labels
        assert result == mock_issue

    @patch("firefighter.raid.service.jira_client")
    def test_create_issue_documentation_request_with_existing_label(self, mock_jira_client):
        """Test create_issue_documentation_request when label already exists (covers branch 112->114)."""
        mock_issue = {"id": "DOC-125"}
        mock_jira_client.create_issue.return_value = mock_issue

        result = create_issue_documentation_request(
            title="Doc with existing label",
            description="Documentation description",
            reporter="reporter_id",
            priority=2,
            labels=["help", "documentation-request"],  # Already has documentation-request
            platform="mobile",
        )

        mock_jira_client.create_issue.assert_called_once()
        call_args = mock_jira_client.create_issue.call_args
        # Should not duplicate the documentation-request label
        expected_labels = ["help", "documentation-request"]
        assert call_args[1]["labels"] == expected_labels
        assert result == mock_issue

    @patch("firefighter.raid.service.jira_client")
    def test_create_issue_internal(self, mock_jira_client):
        """Test create_issue_internal function."""
        mock_issue = {"id": "INT-123"}
        mock_jira_client.create_issue.return_value = mock_issue

        result = create_issue_internal(
            title="Internal Issue",
            description="Internal issue description",
            reporter="reporter_id",
            priority=1,
            labels=["internal"],
            platform="api",
            business_impact="critical",
            team_to_be_routed="backend-team",
            incident_category="infrastructure",
        )

        mock_jira_client.create_issue.assert_called_once()
        call_args = mock_jira_client.create_issue.call_args
        assert call_args[1]["issuetype"] == "Incident"
        assert call_args[1]["summary"] == "Internal Issue"

        assert result == mock_issue

    @patch("firefighter.raid.service.jira_client")
    def test_create_issue_seller(self, mock_jira_client):
        """Test create_issue_seller function."""
        mock_issue = {"id": "SELL-123"}
        mock_jira_client.create_issue.return_value = mock_issue

        result = create_issue_seller(
            title="Seller Issue",
            description="Seller issue description",
            reporter="reporter_id",
            priority=2,
            labels=["seller"],
            platform="web",
            business_impact="medium",
            team_to_be_routed="seller-team",
            incident_category="marketplace",
            seller_contract_id="CONTRACT-123",
            is_key_account=True,
            is_seller_in_golden_list=False,
            zoho_desk_ticket_id="ZOHO-456",
        )

        mock_jira_client.create_issue.assert_called_once()
        call_args = mock_jira_client.create_issue.call_args
        assert call_args[1]["issuetype"] == "Incident"
        assert call_args[1]["summary"] == "Seller Issue"

        assert result == mock_issue


@pytest.mark.django_db
class TestCreateIssueErrorHandling:
    """Test error handling in create_issue functions."""

    @patch("firefighter.raid.service.jira_client")
    def test_create_issue_customer_with_jira_client_returning_none(self, mock_jira_client):
        """Test create_issue_customer when jira_client returns None."""
        mock_jira_client.create_issue.return_value = None

        issue_data = CustomerIssueData(
            priority=1,
            labels=["test"],
            platform="web",
            business_impact="low",
            team_to_be_routed="test-team",
            area="test",
            zendesk_ticket_id="123",
        )

        with pytest.raises(AttributeError):
            create_issue_customer(
                title="Test",
                description="Test description",
                reporter="test_reporter",
                issue_data=issue_data,
            )

    @patch("firefighter.raid.service.jira_client")
    def test_create_issue_customer_with_empty_issue_data(self, mock_jira_client):
        """Test create_issue_customer with minimal issue_data."""
        mock_issue = {"id": "TEST-123"}
        mock_jira_client.create_issue.return_value = mock_issue

        issue_data = CustomerIssueData(
            priority=None,
            labels=None,
            platform="test",
            business_impact=None,
            team_to_be_routed=None,
            area=None,
            zendesk_ticket_id=None,
        )

        result = create_issue_customer(
            title="Minimal Test",
            description="Minimal test description",
            reporter="test_reporter",
            issue_data=issue_data,
        )

        # Should still work with None values
        mock_jira_client.create_issue.assert_called_once()
        assert result == mock_issue



# Generated manually on 2025-08-19 - Step 2: Copy Component data to IncidentCategory

from django.db import migrations


def copy_component_data_to_incident_category(apps, schema_editor):
    """Copy all data from Component to IncidentCategory"""
    Component = apps.get_model("incidents", "Component")
    IncidentCategory = apps.get_model("incidents", "IncidentCategory")

    # Copy all components to incident categories with the same fields
    for component in Component.objects.all():
        IncidentCategory.objects.create(
            id=component.id,  # Keep same UUID
            name=component.name,
            description=component.description,
            order=component.order,
            private=component.private,
            deploy_warning=component.deploy_warning,
            created_at=component.created_at,
            updated_at=component.updated_at,
            group=component.group,
        )


def reverse_copy_component_data_to_incident_category(apps, schema_editor):
    """Reverse operation - copy IncidentCategory back to Component if needed"""
    Component = apps.get_model("incidents", "Component")
    IncidentCategory = apps.get_model("incidents", "IncidentCategory")

    # This would only work if Component table still exists during rollback
    for incident_category in IncidentCategory.objects.all():
        Component.objects.create(
            id=incident_category.id,
            name=incident_category.name,
            description=incident_category.description,
            order=incident_category.order,
            private=incident_category.private,
            deploy_warning=incident_category.deploy_warning,
            created_at=incident_category.created_at,
            updated_at=incident_category.updated_at,
            group=incident_category.group,
        )


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0020_create_incident_category_model"),
    ]

    operations = [
        migrations.RunPython(
            copy_component_data_to_incident_category,
            reverse_copy_component_data_to_incident_category,
        ),
    ]



# Generated by Django 4.2.20 on 2025-03-25 10:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("slack", "0002_usergroup_tag"),
    ]

    operations = [
        migrations.AlterField(
            model_name="usergroup",
            name="tag",
            field=models.CharField(
                blank=True,
                help_text="Used by FireFighter internally to mark special user groups (e.g. @team-secu, @team-incidents...). Must be empty or unique.",
                max_length=80,
            ),
        ),
    ]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.conf import settings
from pagerduty import RestApiV2Client

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

    from httpx import Response

logger = logging.getLogger(__name__)


class PagerdutyClient:
    api_key = settings.PAGERDUTY_API_KEY
    session = RestApiV2Client(api_key, default_from=settings.PAGERDUTY_ACCOUNT_EMAIL)

    def get_schedule_on_call_users(
        self, schedule: str, since: datetime, until: datetime
    ) -> Response:
        if since is not None and until is not None and since > until:
            raise ValueError("Since must be before until")

        return self.session.get(
            f"schedules/{schedule}/users",
            params={"since": str(since), "until": str(until)},
        )

    def get_user_contact_method(self, user_id: str, contact_method_id: str) -> Response:
        return self.session.get(f"users/{user_id}/contact_methods/{contact_method_id}")

    def get_all_services(self) -> Iterator[dict[Any, Any]]:
        return self.session.iter_all("services")

    def get_all_users(self) -> Iterator[dict[Any, Any]]:
        return self.session.iter_all("users", params={"include[]": "contact_methods"})

    def create_incident(
        self,
        title: str,
        pagerduty_id: str,
        details: str,
        incident_key: str,
        conference_url: str,
    ) -> Response:
        return self.session.post(
            url="incidents",
            json={
                "incident": {
                    "type": "incident",
                    "title": title,
                    "service": {"id": pagerduty_id, "type": "service_reference"},
                    "urgency": "high",
                    "incident_key": incident_key,
                    "body": {
                        "type": "incident_body",
                        "details": details,
                    },
                    "conference_bridge": {"conference_url": conference_url},
                }
            },
        )



"""Service for creating and managing Jira post-mortems."""

from __future__ import annotations

import logging
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from django.conf import settings
from django.template.loader import render_to_string

from firefighter.jira_app.client import (
    JiraClient,
    JiraUserDatabaseError,
    JiraUserNotFoundError,
)
from firefighter.jira_app.models import JiraPostMortem

if TYPE_CHECKING:
    from firefighter.incidents.models.incident import Incident
    from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)


class JiraPostMortemService:
    """Service for creating and managing Jira post-mortems."""

    def __init__(self) -> None:
        self.client = JiraClient()
        self.project_key = getattr(settings, "JIRA_POSTMORTEM_PROJECT_KEY", "INCIDENT")
        self.issue_type = getattr(settings, "JIRA_POSTMORTEM_ISSUE_TYPE", "Post-mortem")
        self.ready_status_name = getattr(
            settings, "JIRA_POSTMORTEM_READY_STATUS", "Ready"
        )
        self.field_ids = getattr(
            settings,
            "JIRA_POSTMORTEM_FIELDS",
            {
                "incident_summary": "customfield_12699",
                "timeline": "customfield_12700",
                "root_causes": "customfield_12701",
                "impact": "customfield_12702",
                "mitigation_actions": "customfield_12703",
                "incident_category": "customfield_12369",
            },
        )

    def create_postmortem_for_incident(
        self,
        incident: Incident,
        created_by: User | None = None,
    ) -> JiraPostMortem:
        """Create a Jira post-mortem for an incident.

        Args:
            incident: Incident to create post-mortem for
            created_by: User creating the post-mortem

        Returns:
            JiraPostMortem instance

        Raises:
            ValueError: If incident already has a Jira post-mortem
            JiraAPIError: If Jira API call fails
        """
        if hasattr(incident, "jira_postmortem_for"):
            error_msg = f"Incident #{incident.id} already has a Jira post-mortem"
            raise ValueError(error_msg)

        logger.info(f"Creating Jira post-mortem for incident #{incident.id}")

        # Prefetch incident updates and jira_ticket for timeline and parent link
        from firefighter.incidents.models.incident import Incident

        incident = (
            Incident.objects.select_related("priority", "environment", "jira_ticket")
            .prefetch_related("incidentupdate_set")
            .get(pk=incident.pk)
        )

        # Generate content from templates
        fields = self._generate_issue_fields(incident)

        # Get parent issue key from RAID Jira ticket if available
        parent_issue_key = None
        if hasattr(incident, "jira_ticket") and incident.jira_ticket:
            parent_issue_key = incident.jira_ticket.key

        # Create Jira issue with optional parent link
        jira_issue = self.client.create_postmortem_issue(
            project_key=self.project_key,
            issue_type=self.issue_type,
            fields=fields,
            parent_issue_key=parent_issue_key,
        )

        # Assign to incident commander if available
        commander = (
            incident.roles_set.select_related("user__jira_user", "role_type")
            .filter(role_type__slug="commander")
            .first()
        )
        if commander:
            jira_user = getattr(commander.user, "jira_user", None)
            if jira_user is None:
                try:
                    jira_user = self.client.get_jira_user_from_user(commander.user)
                except (JiraUserNotFoundError, JiraUserDatabaseError) as exc:
                    logger.warning(
                        "Unable to fetch Jira user for commander %s: %s",
                        commander.user_id,
                        exc,
                    )
            if jira_user is not None:
                assigned = self.client.assign_issue(
                    issue_key=jira_issue["key"],
                    account_id=jira_user.id,
                )
                if assigned:
                    logger.info(
                        "Assigned post-mortem %s to commander %s",
                        jira_issue["key"],
                        commander.user.username,
                    )

        # Create JiraPostMortem record
        jira_postmortem = JiraPostMortem.objects.create(
            incident=incident,
            jira_issue_key=jira_issue["key"],
            jira_issue_id=jira_issue["id"],
            created_by=created_by,
        )

        logger.info(
            f"Created Jira post-mortem {jira_postmortem.jira_issue_key} "
            f"for incident #{incident.id}"
        )

        return jira_postmortem

    def is_postmortem_ready(self, jira_postmortem: JiraPostMortem) -> tuple[bool, str]:
        """Check if the Jira post-mortem issue is in the Ready status.

        Returns:
            Tuple of (is_ready, current_status_name)
        """
        issue = self.client.jira.issue(jira_postmortem.jira_issue_id)
        status_name: str = getattr(issue.fields.status, "name", "")
        return status_name == self.ready_status_name, status_name

    def _generate_issue_fields(
        self, incident: Incident
    ) -> dict[str, str | dict[str, str] | list[dict[str, str]]]:
        """Generate Jira issue fields from incident data.

        Args:
            incident: Incident to generate fields for

        Returns:
            Dictionary of Jira field IDs to values
        """
        # Generate summary (standard field)
        env = getattr(settings, "ENV", "dev")
        topic_prefix = "" if env in {"support", "prod"} else f"[IGNORE - TEST {env}] "
        summary = (
            f"{topic_prefix}#{incident.slack_channel_name} "
            f"({incident.priority.name}) {incident.title}"
        )

        # Generate content from Django templates (Jira Wiki Markup)
        context = {
            "incident": incident,
            "priority": incident.priority,
            "created_at": incident.created_at,
            "components": [],  # No component relationship available
        }

        incident_summary = render_to_string(
            "jira/postmortem/incident_summary.txt",
            context,
        )

        timeline = render_to_string(
            "jira/postmortem/timeline.txt",
            context,
        )

        impact = render_to_string(
            "jira/postmortem/impact.txt",
            context,
        )

        mitigation_actions = render_to_string(
            "jira/postmortem/mitigation_actions.txt",
            context,
        )

        # Optional: root causes (editable placeholder for manual completion)
        root_causes = render_to_string(
            "jira/postmortem/root_causes.txt",
            context,
        )

        # Build field mapping
        fields: dict[str, str | dict[str, str] | list[dict[str, str]]] = {
            "summary": summary,
            self.field_ids["incident_summary"]: incident_summary,
            self.field_ids["timeline"]: timeline,
            self.field_ids["impact"]: impact,
            self.field_ids["mitigation_actions"]: mitigation_actions,
        }

        due_date = self._add_business_days(incident.created_at, 40).date()
        fields["duedate"] = due_date.isoformat()

        # Add optional fields if not empty
        if root_causes.strip():
            fields[self.field_ids["root_causes"]] = root_causes

        # Add incident category if available
        if incident.incident_category:
            # Jira select field requires dict with value key
            category_field_id = self.field_ids["incident_category"]
            fields[category_field_id] = {"value": incident.incident_category.name}

        # Replicate custom fields from incident ticket to post-mortem
        self._add_replicated_custom_fields(incident, fields)

        return fields

    def _add_replicated_custom_fields(
        self,
        incident: Incident,
        fields: dict[str, str | dict[str, str] | list[dict[str, str]]],
    ) -> None:
        """Replicate custom fields from incident ticket to post-mortem.

        Replicates the following fields from the incident ticket:
        - Priority (customfield_11064)
        - Affected environments (customfield_11049)
        - Zoho desk ticket (customfield_10896)
        - Zendesk ticket (customfield_10895)
        - Seller Contract ID (customfield_10908)
        - Platforms (customfield_10201) - first platform from list
        - Business Impact (customfield_10936)

        Args:
            incident: Incident to extract fields from
            fields: Dictionary to add fields to (modified in place)
        """
        custom_fields = incident.custom_fields or {}

        # Priority - customfield_11064 (option field)
        if incident.priority:
            priority_value = str(incident.priority.value)
            fields["customfield_11064"] = {"value": priority_value}

        # Affected environments - customfield_11049 (array field)
        environments = custom_fields.get("environments", [])
        if environments:
            fields["customfield_11049"] = [{"value": env} for env in environments]

        # Zendesk ticket - customfield_10895 (string field)
        zendesk_ticket_id = custom_fields.get("zendesk_ticket_id")
        if zendesk_ticket_id:
            fields["customfield_10895"] = str(zendesk_ticket_id)

        # Zoho desk ticket - customfield_10896 (string field)
        zoho_desk_ticket_id = custom_fields.get("zoho_desk_ticket_id")
        if zoho_desk_ticket_id:
            fields["customfield_10896"] = str(zoho_desk_ticket_id)

        # Seller Contract ID - customfield_10908 (string field)
        seller_contract_id = custom_fields.get("seller_contract_id")
        if seller_contract_id:
            fields["customfield_10908"] = str(seller_contract_id)

        # Platform - customfield_10201 (option field)
        platforms = custom_fields.get("platforms", [])
        if platforms:
            # Extract first platform and remove "platform-" prefix if present
            platform = platforms[0] if isinstance(platforms, list) else platforms
            platform_value = platform.replace("platform-", "") if isinstance(platform, str) else platform
            fields["customfield_10201"] = {"value": platform_value}

        # Business Impact - customfield_10936 (option field)
        # Business impact is stored in the Jira ticket, not in custom_fields
        if hasattr(incident, "jira_ticket") and incident.jira_ticket:
            business_impact = incident.jira_ticket.business_impact
            if business_impact and business_impact not in {"", "N/A"}:
                fields["customfield_10936"] = {"value": business_impact}

    @staticmethod
    def _add_business_days(start: datetime, days: int) -> datetime:
        current = start
        added = 0
        while added < days:
            current += timedelta(days=1)
            if current.weekday() < 5:
                added += 1
        return current


# Singleton instance
jira_postmortem_service = JiraPostMortemService()



"""Django management command to switch Slack user IDs for test environment.

This command:
1. Fetches all users from the test Slack workspace
2. Generates a mapping file based on email addresses
3. Updates the database with test Slack user IDs
"""

from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import Any

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

from firefighter.incidents.models.user import User
from firefighter.slack.models.user import SlackUser

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Generate and apply Slack user mapping for test environment"

    def add_arguments(self, parser: Any) -> None:
        parser.add_argument(
            "--generate-only",
            action="store_true",
            help="Only generate the mapping file, don't apply changes",
        )
        parser.add_argument(
            "--mapping-file",
            type=str,
            default="slack_test_mapping.json",
            help="Path to the mapping file (default: slack_test_mapping.json)",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Show what would be changed without making actual changes",
        )
        parser.add_argument(
            "--restore",
            action="store_true",
            help="Restore original Slack IDs from mapping file",
        )

    def _raise_mapping_file_not_found(self, mapping_file: Path) -> None:
        """Raise error when mapping file is not found."""
        error_msg = f"Mapping file {mapping_file} not found"
        raise CommandError(error_msg)

    def _raise_slack_token_not_configured(self) -> None:
        """Raise error when Slack token is not configured."""
        raise CommandError("SLACK_BOT_TOKEN not configured")

    def handle(self, *args: Any, **options: Any) -> None:
        """Main command handler."""
        mapping_file = Path(options["mapping_file"])

        try:
            if options["restore"]:
                # Restore original Slack IDs from mapping file
                if not mapping_file.exists():
                    self._raise_mapping_file_not_found(mapping_file)

                self.stdout.write("📄 Loading mapping from file...")
                slack_mapping = self._load_mapping_file(mapping_file)

                self.stdout.write("🔄 Restoring original Slack IDs...")
                updated_count = self._restore_mapping(slack_mapping, dry_run=options["dry_run"])

                if options["dry_run"]:
                    self.stdout.write(
                        self.style.WARNING(f"🔍 DRY RUN: Would restore {updated_count} users")
                    )
                else:
                    self.stdout.write(
                        self.style.SUCCESS(f"✅ Restored {updated_count} users to original Slack IDs")
                    )
                return

            if not settings.SLACK_BOT_TOKEN:
                self._raise_slack_token_not_configured()

            # Step 1: Generate mapping by querying test Slack workspace
            self.stdout.write("🔍 Fetching users from test Slack workspace...")
            slack_mapping = self._generate_slack_mapping()

            # Step 2: Save mapping to file
            self._save_mapping_file(slack_mapping, mapping_file)
            self.stdout.write(
                self.style.SUCCESS(f"✅ Mapping file saved to {mapping_file}")
            )

            if options["generate_only"]:
                self.stdout.write("📄 Generation complete. Use --apply to update database.")
                return

            # Step 3: Apply mapping to database
            self.stdout.write("📝 Applying mapping to database...")
            updated_count = self._apply_mapping(slack_mapping, dry_run=options["dry_run"])

            if options["dry_run"]:
                self.stdout.write(
                    self.style.WARNING(f"🔍 DRY RUN: Would update {updated_count} users")
                )
            else:
                self.stdout.write(
                    self.style.SUCCESS(f"✅ Updated {updated_count} users with test Slack IDs")
                )

        except SlackApiError as e:
            error_msg = f"Slack API error: {e.response['error']}"
            raise CommandError(error_msg) from e
        except Exception as e:
            error_msg = f"Error: {e}"
            raise CommandError(error_msg) from e

    def _generate_slack_mapping(self) -> dict[str, dict[str, Any]]:
        """Generate mapping by querying test Slack workspace."""
        client = WebClient(token=settings.SLACK_BOT_TOKEN)

        # Get all users from test Slack workspace
        try:
            response = client.users_list()
            slack_users = response["members"]
        except SlackApiError as e:
            error_msg = f"Failed to fetch Slack users: {e.response['error']}"
            raise CommandError(error_msg) from e

        # Create mapping: email -> test_slack_id
        slack_email_to_id = {}
        for slack_user in slack_users:
            if slack_user.get("deleted") or slack_user.get("is_bot"):
                continue

            profile = slack_user.get("profile", {})
            email = profile.get("email")
            if email:
                slack_email_to_id[email.lower()] = slack_user["id"]

        self.stdout.write(f"📧 Found {len(slack_email_to_id)} users with emails in test Slack")

        # Get all users from database
        db_users = User.objects.exclude(email__isnull=True).exclude(email="")

        # Generate mapping for users that exist in both places
        mapping = {}
        matched_count = 0
        not_found_count = 0

        for db_user in db_users:
            if db_user.email and db_user.email.lower() in slack_email_to_id:
                test_slack_id = slack_email_to_id[db_user.email.lower()]
                # Get current slack_id from SlackUser relation
                current_slack_id = None
                if hasattr(db_user, "slack_user") and db_user.slack_user:
                    current_slack_id = db_user.slack_user.slack_id

                mapping[db_user.email] = {
                    "original_slack_id": current_slack_id,
                    "test_slack_id": test_slack_id,
                    "name": f"{db_user.first_name} {db_user.last_name}".strip(),
                }
                matched_count += 1
            else:
                not_found_count += 1

        self.stdout.write(f"🎯 Matched {matched_count} users between database and test Slack")
        if not_found_count > 0:
            self.stdout.write(f"⚠️  {not_found_count} database users not found in test Slack")
        return mapping

    def _save_mapping_file(self, mapping: dict[str, Any], file_path: Path) -> None:
        """Save mapping to JSON file."""
        with file_path.open("w", encoding="utf-8") as f:
            json.dump(mapping, f, indent=2, ensure_ascii=False)

    def _load_mapping_file(self, file_path: Path) -> dict[str, Any]:
        """Load mapping from JSON file."""
        with file_path.open(encoding="utf-8") as f:
            return json.load(f)

    def _apply_mapping(self, mapping: dict[str, Any], *, dry_run: bool = False) -> int:
        """Apply the mapping to update database users."""
        updated_count = 0

        for email, user_data in mapping.items():
            test_slack_id = user_data["test_slack_id"]

            try:
                user = User.objects.get(email=email)

                # Get current slack_id from SlackUser relation
                current_slack_id = None
                if hasattr(user, "slack_user") and user.slack_user:
                    current_slack_id = user.slack_user.slack_id

                if current_slack_id == test_slack_id:
                    continue  # Already has the test ID

                if dry_run:
                    self.stdout.write(
                        f"🔄 Would update {email}: {current_slack_id} -> {test_slack_id}"
                    )
                elif hasattr(user, "slack_user") and user.slack_user:
                    user.slack_user.slack_id = test_slack_id
                    user.slack_user.save(update_fields=["slack_id"])
                    self.stdout.write(f"✅ Updated {email}: {current_slack_id} -> {test_slack_id}")
                else:
                    # Create new SlackUser if it doesn't exist
                    SlackUser.objects.create(user=user, slack_id=test_slack_id)
                    self.stdout.write(f"✅ Created SlackUser for {email}: -> {test_slack_id}")

                updated_count += 1

            except User.DoesNotExist:
                self.stdout.write(f"❌ User not found in database: {email}")
            except (SlackUser.DoesNotExist, ValueError) as e:
                self.stdout.write(f"❌ Error updating {email}: {e}")

        return updated_count

    def _restore_mapping(self, mapping: dict[str, Any], *, dry_run: bool = False) -> int:
        """Restore original Slack IDs from mapping."""
        updated_count = 0

        for email, user_data in mapping.items():
            original_slack_id = user_data["original_slack_id"]

            try:
                user = User.objects.get(email=email)

                # Get current slack_id from SlackUser relation
                current_slack_id = None
                if hasattr(user, "slack_user") and user.slack_user:
                    current_slack_id = user.slack_user.slack_id

                if current_slack_id == original_slack_id:
                    continue  # Already has the original ID

                if dry_run:
                    self.stdout.write(
                        f"🔄 Would restore {email}: {current_slack_id} -> {original_slack_id}"
                    )
                elif hasattr(user, "slack_user") and user.slack_user:
                    if original_slack_id:
                        user.slack_user.slack_id = original_slack_id
                        user.slack_user.save(update_fields=["slack_id"])
                        self.stdout.write(f"✅ Restored {email}: {current_slack_id} -> {original_slack_id}")
                    else:
                        # If original was None, delete the SlackUser
                        user.slack_user.delete()
                        self.stdout.write(f"✅ Removed SlackUser for {email}: {current_slack_id} -> None")
                else:
                    self.stdout.write(f"⚠️  User {email} has no SlackUser to restore")
                    continue

                updated_count += 1

            except User.DoesNotExist:
                self.stdout.write(f"❌ User not found in database: {email}")
            except (SlackUser.DoesNotExist, ValueError) as e:
                self.stdout.write(f"❌ Error restoring {email}: {e}")

        return updated_count



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django import forms

from firefighter.incidents.forms.select_impact import SelectImpactForm
from firefighter.incidents.forms.utils import GroupedModelChoiceField
from firefighter.incidents.models import Environment, IncidentCategory, Priority
from firefighter.incidents.models.incident import Incident
from firefighter.incidents.signals import create_incident_conversation

if TYPE_CHECKING:
    from firefighter.incidents.models.impact import ImpactLevel
    from firefighter.incidents.models.user import User


def initial_environments() -> Environment:
    return Environment.objects.get(default=True)


def initial_priority() -> Priority:
    return Priority.objects.get(default=True)


class CreateIncidentFormBase(forms.Form):
    def trigger_incident_workflow(
        self,
        creator: User,
        impacts_data: dict[str, ImpactLevel],
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        raise NotImplementedError


class CreateIncidentForm(CreateIncidentFormBase):
    title = forms.CharField(
        label="Title",
        max_length=128,
        min_length=10,
        widget=forms.TextInput(attrs={"placeholder": "What's going on?"}),
    )
    description = forms.CharField(
        label="Summary",
        widget=forms.Textarea(
            attrs={
                "placeholder": "Help people responding to the incident. This will be posted to #tech-incidents and on our internal status page.\nThis description can be edited later."
            }
        ),
        min_length=10,
        max_length=1200,
    )

    incident_category = GroupedModelChoiceField(
        choices_groupby="group",
        label="Incident category",
        queryset=(
            IncidentCategory.objects.all()
            .select_related("group")
            .order_by(
                "group__order",
                "name",
            )
        ),
    )
    priority = forms.ModelChoiceField(
        label="Priority",
        queryset=Priority.objects.filter(enabled_create=True),
        initial=initial_priority,
    )
    environment = forms.ModelChoiceField(
        label="Environment",
        queryset=Environment.objects.all(),
        initial=initial_environments,
    )

    def trigger_incident_workflow(
        self,
        creator: User,
        impacts_data: dict[str, ImpactLevel],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        incident = Incident.objects.declare(created_by=creator, **self.cleaned_data)
        impacts_form = SelectImpactForm(impacts_data)

        impacts_form.save(incident=incident)
        create_incident_conversation.send(
            "create_incident_form",
            incident=incident,
        )



# Generated by Django 4.2.21 on 2025-10-13 13:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0028_add_closure_reason_constraint"),
    ]

    operations = [
        migrations.AddField(
            model_name="incident",
            name="custom_fields",
            field=models.JSONField(
                blank=True,
                default=dict,
                help_text="Custom fields for incident (zendesk_ticket_id, seller_contract_id, etc.)",
            ),
        ),
    ]



from __future__ import annotations

import logging

from celery import shared_task

from firefighter.pagerduty.models import PagerDutyTeam, PagerDutyUser
from firefighter.pagerduty.service import pagerduty_service
from firefighter.slack.models import SlackUser

logger = logging.getLogger(__name__)


@shared_task(name="pagerduty.fetch_users")
def fetch_users(*, delete_stale_user: bool = True) -> None:
    """Celery task to fetch PagerDuty users and save them in the database."""
    fetched_users_id = []
    for user in pagerduty_service.client.get_all_users():
        logger.debug(user)
        fetched_users_id.append(user["id"])
        main_user = SlackUser.objects.upsert_by_email(user["email"])
        if main_user is None:
            logger.warning("Could not find user with email %s", user["email"])
            continue

        phone_number = pagerduty_service.get_phone_number_from_body(user)
        if phone_number is None:
            phone_number = ""

        pd_user, _ = PagerDutyUser.objects.update_or_create(
            pagerduty_id=user["id"],
            defaults={
                "name": user["name"],
                "user": main_user,
                "phone_number": phone_number,
                "pagerduty_url": user["html_url"][:256],
                "pagerduty_api_url": user["self"][:256],
            },
        )
        pd_teams = user.get("teams", [])
        pd_teams_models: list[PagerDutyTeam] = []
        for pd_team in pd_teams:
            pd_team_model, _ = PagerDutyTeam.objects.update_or_create(
                pagerduty_id=pd_team["id"],
                defaults={
                    "name": pd_team["summary"],
                    "pagerduty_url": pd_team["html_url"][:256],
                    "pagerduty_api_url": pd_team["self"][:256],
                },
            )
            pd_teams_models.append(pd_team_model)

        pd_user.teams.set(pd_teams_models)

    # Check that we don't have stale users
    if len(fetched_users_id) != PagerDutyUser.objects.count():
        stale_user_ids = PagerDutyUser.objects.exclude(
            pagerduty_id__in=fetched_users_id
        ).values_list("pagerduty_id", flat=True)
        logger.info(f"Stale Pagerduty users found {list(stale_user_ids)}.")

        if delete_stale_user:
            nb_deleted, _ = PagerDutyUser.objects.filter(
                pagerduty_id__in=stale_user_ids
            ).delete()
            logger.info(f"Deleted {nb_deleted} stale PagerDuty users.")



from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import Any, NotRequired, Required, TypedDict

from django.urls import reverse
from django_components import component


class _Format(TypedDict):
    label: str
    url: str
    fields: str


class Data(TypedDict):
    default_format: Mapping[str, Any]  # _Format
    formats: Sequence[Mapping[str, Any]]  # list[_Format]


Args = tuple[str]


class Kwargs(TypedDict, total=False):
    base_url: Required[str]
    default_fmt: NotRequired[tuple[str, str | None, str | None]]


@component.register("export_button")
class ExportButton(component.Component):
    template_name = "export_button/export_button.html"

    def get_context_data(
        self,
        base_url: str,
        default_fmt: tuple[str, str | None, str | None] | None = None,
        **kwargs: Any,
    ) -> Data:
        default_fmt = default_fmt or ("csv", None, None)
        fmts: list[tuple[str, str | None, str | None]] = [
            ("json", None, None),
            ("tsv", None, None),
            ("csv", "__all__", "(Full)"),
            ("tsv", "__all__", "(Full)"),
        ]

        default_format = self.make_fmt(default_fmt, base_url)
        formats = [self.make_fmt(fmt, base_url) for fmt in fmts]

        return Data(default_format=default_format, formats=formats)

    @staticmethod
    def make_fmt(
        format_: tuple[str, str | None, str | None], base_reverse: str
    ) -> _Format:
        return {
            "label": f"Export {format_[0].upper()}{' ' + format_[2] if format_[2] else ''}",
            "url": f"{reverse(base_reverse, args=[format_[0]])}",
            "fields": format_[1] or "",
        }



"""Utilities for modal handling to avoid circular imports."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.forms.update_status import UpdateStatusForm
from firefighter.slack.views.modals.closure_reason import modal_closure_reason

if TYPE_CHECKING:
    from slack_sdk.models.views import View

    from firefighter.incidents.models.incident import Incident


def get_close_modal_view(body: dict[str, Any], incident: Incident, **kwargs: Any) -> View | None:
    """Get the appropriate modal view for closing an incident.

    This function determines whether to show the closure reason modal
    or delegate to the normal close modal.
    """
    # Check if closure reason is required
    if UpdateStatusForm.requires_closure_reason(incident, IncidentStatus.CLOSED):
        return modal_closure_reason.build_modal_fn(body, incident, **kwargs)

    # Return None to indicate normal close modal should be used
    return None


def handle_close_modal_callback(ack: Any, body: dict[str, Any], incident: Incident, user: Any) -> bool | None:
    """Handle modal callback, delegating to closure reason modal if needed."""
    # Check if this is a closure reason modal callback
    if body.get("view", {}).get("callback_id") == "incident_closure_reason":
        return modal_closure_reason.handle_modal_fn(ack, body, incident, user)

    # Return None to indicate normal handling should continue
    return None


def handle_update_status_close_request(ack: Any, body: dict[str, Any], incident: Incident, target_status: IncidentStatus) -> bool:
    """Handle update status request to close incident, showing reason modal if needed.

    Returns True if the request was handled (reason modal shown), False otherwise.
    """
    if (target_status == IncidentStatus.CLOSED and
        UpdateStatusForm.requires_closure_reason(incident, target_status)):
        # Show closure reason modal instead
        ack(response_action="push", view=modal_closure_reason.build_modal_fn(body, incident))
        return True

    return False



# Generated manually on 2025-08-19 - Step 6: Make incident_category field required

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0024_remove_component_fields_and_model"),
        ("slack", "0007_remove_components_fields"),
    ]

    operations = [
        # Make incident_category field required (not null)
        migrations.AlterField(
            model_name="incident",
            name="incident_category",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to="incidents.incidentcategory",
            ),
        ),
    ]



from __future__ import annotations

import functools
from calendar import monthrange
from datetime import date, datetime, time, timedelta
from typing import TYPE_CHECKING

from django.conf import settings
from django.core.exceptions import BadRequest
from django.utils import timezone

if TYPE_CHECKING:
    from dateparser.date import DateDataParser

TZ = timezone.get_current_timezone()


@functools.cache
def get_ddp() -> DateDataParser:
    from dateparser.date import DateDataParser

    return DateDataParser(
        locales=["en"],
        settings={
            "RETURN_TIME_AS_PERIOD": True,
            "DATE_ORDER": "YMD",
            "TIMEZONE": settings.TIME_ZONE,
            "RETURN_AS_TIMEZONE_AWARE": True,
        },
    )


def get_current_quarter() -> tuple[int, int]:
    year, week, _weekday = (timezone.localtime(timezone.now()).date()).isocalendar()
    current_week_monday = date.fromisocalendar(year=year, week=week, day=1)
    current_week_year, current_week_number, _ = current_week_monday.isocalendar()
    current_quarter = get_quarter_from_week(current_week_number)
    return current_week_year, current_quarter


def get_last_quarter() -> tuple[int, int]:
    current_year, current_quarter = get_current_quarter()
    if current_quarter == 1:
        return current_year - 1, 4
    return current_year, current_quarter - 1


def get_quarter_from_week(week_number: str | int) -> int:
    week_number = int(week_number)
    if 0 < week_number <= 13:
        return 1
    if 13 < week_number <= 26:
        return 2
    if 26 < week_number <= 39:
        return 3
    if 39 < week_number <= 53:
        return 4
    raise BadRequest("Invalid week number.")


def get_bounds_from_calendar_year(year_value: int) -> tuple[datetime, datetime, str]:
    """Calendar Year."""
    lower_bound = datetime(year_value, 1, 1, tzinfo=TZ)
    upper_bound = datetime(year_value, 12, 31, 23, 59, 59, 999999, tzinfo=TZ)

    return (
        lower_bound,
        upper_bound,
        f"{year_value}",
    )


def get_bounds_from_year(year_value: int) -> tuple[datetime, datetime, str]:
    """ISO Year."""
    lower_bound = datetime.fromisocalendar(year_value, 1, 1)
    upper_week_nb = date(year_value, 12, 28).isocalendar()[1]
    upper_bound = datetime.combine(
        date.fromisocalendar(year=year_value, week=upper_week_nb, day=7),
        time(23, 59, 59, 999999),
    )

    return *make_bounds_timezone_aware(lower_bound, upper_bound), f"ISO{year_value}"


def get_bounds_from_week(
    year_value: int, week_value: int
) -> tuple[datetime, datetime, str]:
    lower_bound = datetime.fromisocalendar(year_value, week_value, 1).astimezone(TZ)
    upper_bound = datetime.combine(
        date.fromisocalendar(year_value, week_value, 7),
        time(23, 59, 59, 999999),
        tzinfo=TZ,
    )

    return lower_bound, upper_bound, f"{year_value}-W{week_value}"


def get_bounds_from_quarter(
    year_nb: int, quarter_nb: int
) -> tuple[datetime, datetime, str]:
    lower_week_nb = quarter_nb * 13 - 12
    lower_bound = datetime.fromisocalendar(year_nb, lower_week_nb, 1)
    if quarter_nb != 4:
        upper_bound = datetime.fromisocalendar(
            year_nb, lower_week_nb + 13, 1
        ) - timedelta(microseconds=1)
    else:
        # The last week of ISO year always contains the 28th december. It can be W52 or W53.
        upper_week_nb = date(year_nb, 12, 28).isocalendar()[1]
        upper_bound = datetime.combine(
            date.fromisocalendar(year=year_nb, week=upper_week_nb, day=7),
            time(23, 59, 59, 999999),
        )

    return (
        *make_bounds_timezone_aware(lower_bound, upper_bound),
        f"{year_nb}-Q{quarter_nb}",
    )


def get_day_date_range(date_: date) -> slice:
    return slice(
        datetime(date_.year, date_.month, date_.day, 0, 0, 0, tzinfo=TZ),
        datetime(date_.year, date_.month, date_.day, 23, 59, 59, 999999, tzinfo=TZ),
    )


def get_month_date_range(date_: date) -> slice:
    return slice(
        datetime(date_.year, date_.month, 1, 0, 0, 0, tzinfo=TZ),
        datetime(
            date_.year,
            date_.month,
            monthrange(date_.year, date_.month)[1],
            23,
            59,
            59,
            999999,
            tzinfo=TZ,
        ),
    )


def make_bounds_timezone_aware(
    lower_bound: datetime, upper_bound: datetime
) -> tuple[datetime, datetime]:
    if settings.USE_TZ:
        tz = timezone.get_current_timezone()
        lower_bound = timezone.make_aware(lower_bound, tz)
        upper_bound = timezone.make_aware(upper_bound, tz)
    return lower_bound, upper_bound


def make_timezone_aware(lower_bound: datetime) -> datetime:
    if settings.USE_TZ:
        tz = timezone.get_current_timezone()
        lower_bound = timezone.make_aware(lower_bound, tz)
    return lower_bound


def get_biggest_date(a: datetime | None, b: datetime) -> datetime:
    if not a:
        return b
    if not b:
        return a
    if a < b:
        return b
    return a


def parse_date_natural(value: str) -> slice | None:
    # TODO Reuse the DateDataParser and tune the config

    date_date = get_ddp().get_date_data(value)
    if date_date.date_obj:
        date_ = date_date.date_obj
        if date_date.period == "month":
            return get_month_date_range(date_)
        if date_date.period == "day":
            return get_day_date_range(date_)
        if date_date.period == "year":
            a = get_bounds_from_year(date_.year)
            return slice(a[0], a[1])
        if date_date.period == "week":
            a = get_bounds_from_week(date_.year, date_.isocalendar()[1])
            return slice(a[0], a[1])
        return slice(date_, date_)
    return None



from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from django.core.management.base import BaseCommand, CommandError, CommandParser

from firefighter.firefighter.celery_client import app as celery_app

if TYPE_CHECKING:
    from celery.app.task import Task

    Task.__class_getitem__ = classmethod(lambda cls, *args, **kwargs: cls)  # type: ignore[attr-defined]


class Command(BaseCommand):
    help = "Run a celery task from the command line."

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument("task_name", type=str, choices=celery_app.tasks.keys())

    def handle(self, *args: Any, **options: Any) -> None:
        tasks: dict[str, Task[Any, Any]] = cast(
            "dict[str, Task[Any, Any]]", celery_app.tasks
        )
        task_name = options["task_name"]

        if task_name in tasks:
            self.stdout.write(self.style.SUCCESS(f"Running task {task_name}"))
            task: Task[Any, Any] = tasks[task_name]
            task()
            self.stdout.write(self.style.SUCCESS(f"Successfully ran task {task_name}"))
        else:
            sep = "\n - "
            err_msg = f'Task "{task_name}" does not exist. Available tasks are:\n - {sep.join(sorted(tasks))}'
            raise CommandError(err_msg)



"""Test the status workflow logic for update status form."""
from __future__ import annotations

import pytest
from django.apps import apps
from django.test import TestCase

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.factories import IncidentFactory
from firefighter.incidents.forms.update_status import UpdateStatusForm
from firefighter.incidents.models import Environment, Priority


@pytest.mark.django_db
class TestUpdateStatusWorkflow(TestCase):
    """Test that status choices respect workflow rules."""

    @staticmethod
    def _get_available_statuses(status_choices: dict[str, str]) -> list[IncidentStatus]:
        """Convert status_choices dict (string keys) to list of IncidentStatus enums."""
        return [IncidentStatus(int(k)) for k in status_choices]

    def test_p1_p2_cannot_skip_postmortem(self):
        """P1/P2 incidents cannot go directly from Mitigated to Closed."""

        # Get existing P1 or P2 priority that needs post-mortem
        p1_priority = Priority.objects.filter(
            value=1, needs_postmortem=True
        ).first()

        if not p1_priority:
            # If P1 doesn't exist, try P2
            p1_priority = Priority.objects.filter(
                value=2, needs_postmortem=True
            ).first()

        if not p1_priority:
            # Skip test if no priority needing postmortem exists
            pytest.skip("No P1/P2 priority with needs_postmortem=True found in database")

        # Get PRD environment (required for needs_postmortem to be True)
        prd_env = Environment.objects.filter(value="PRD").first()
        if not prd_env:
            pytest.skip("No PRD environment found in database")

        # Create an incident with P1/P2 priority in Mitigated status with PRD env
        incident = IncidentFactory.create(
            priority=p1_priority,
            environment=prd_env,
            _status=IncidentStatus.MITIGATED,
        )

        # Note: incident.needs_postmortem also requires firefighter.confluence to be installed
        # For testing, we'll check the underlying conditions directly
        if apps.is_installed("firefighter.confluence"):
            assert incident.needs_postmortem, "Incident should need postmortem (P1/P2 + PRD)"

        # Create form with the incident
        form = UpdateStatusForm(incident=incident)

        # Check that CLOSED is not in the choices
        status_choices = dict(form.fields["status"].choices)
        available_statuses = self._get_available_statuses(status_choices)
        assert IncidentStatus.CLOSED not in available_statuses
        assert IncidentStatus.POST_MORTEM in available_statuses

    def test_p1_p2_can_close_from_postmortem(self):
        """P1/P2 incidents can go from Post-mortem to Closed."""

        # Get existing P1 or P2 priority that needs post-mortem
        p2_priority = Priority.objects.filter(
            value__in=[1, 2], needs_postmortem=True
        ).first()

        if not p2_priority:
            pytest.skip("No P1/P2 priority with needs_postmortem=True found in database")

        # Get PRD environment
        prd_env = Environment.objects.filter(value="PRD").first()
        if not prd_env:
            pytest.skip("No PRD environment found in database")

        # Create an incident with P1/P2 priority in Post-mortem status with PRD env
        incident = IncidentFactory.create(
            priority=p2_priority,
            environment=prd_env,
            _status=IncidentStatus.POST_MORTEM,
        )

        # Create form with the incident
        form = UpdateStatusForm(incident=incident)

        # Check that CLOSED is in the choices
        status_choices = dict(form.fields["status"].choices)
        available_statuses = self._get_available_statuses(status_choices)
        assert IncidentStatus.CLOSED in available_statuses

    def test_p3_plus_can_skip_postmortem(self):
        """P3+ incidents can go directly from Mitigated to Closed and should not have post-mortem."""
        # Get existing P3+ priority that doesn't need post-mortem
        p3_priority = Priority.objects.filter(
            value__gte=3, needs_postmortem=False
        ).first()

        if not p3_priority:
            # If no P3+ without postmortem, skip the test
            pytest.skip("No P3+ priority with needs_postmortem=False found in database")

        # Create an incident with P3+ priority in Mitigated status
        incident = IncidentFactory.create(
            priority=p3_priority,
            _status=IncidentStatus.MITIGATED,
        )

        # Create form with the incident
        form = UpdateStatusForm(incident=incident)

        # Check that CLOSED is in the choices for P3+ but POST_MORTEM is NOT
        status_choices = dict(form.fields["status"].choices)
        available_statuses = self._get_available_statuses(status_choices)
        assert IncidentStatus.CLOSED in available_statuses
        assert IncidentStatus.POST_MORTEM not in available_statuses, "P3+ incidents should not have post-mortem status available"

    def test_p1_p2_cannot_skip_postmortem_from_mitigating(self):
        """P1/P2 incidents cannot go directly from Mitigating to Closed."""

        # Get existing P1 or P2 priority that needs post-mortem
        p1_priority = Priority.objects.filter(
            value__in=[1, 2], needs_postmortem=True
        ).first()

        if not p1_priority:
            pytest.skip("No P1/P2 priority with needs_postmortem=True found in database")

        # Get PRD environment (required for needs_postmortem to be True)
        prd_env = Environment.objects.filter(value="PRD").first()
        if not prd_env:
            pytest.skip("No PRD environment found in database")

        # Create an incident with P1/P2 priority in MITIGATING (Mitigating) status with PRD env
        incident = IncidentFactory.create(
            priority=p1_priority,
            environment=prd_env,
            _status=IncidentStatus.MITIGATING,  # Test from MITIGATING (Mitigating), not MITIGATED
        )

        # Create form with the incident
        form = UpdateStatusForm(incident=incident)

        # Check that CLOSED and POST_MORTEM are not in the choices from MITIGATING status
        status_choices = dict(form.fields["status"].choices)
        available_statuses = self._get_available_statuses(status_choices)
        assert IncidentStatus.CLOSED not in available_statuses
        assert IncidentStatus.POST_MORTEM not in available_statuses, "P1/P2 should not be able to go to post-mortem from Mitigating"

    def test_p3_can_skip_from_investigating_to_closed(self):
        """P3+ incidents can go directly from Investigating to Closed."""
        # Get existing P3+ priority that doesn't need post-mortem
        p3_priority = Priority.objects.filter(
            value__gte=3, needs_postmortem=False
        ).first()

        if not p3_priority:
            pytest.skip("No P3+ priority with needs_postmortem=False found in database")

        # Create an incident with P3+ priority in INVESTIGATING status
        incident = IncidentFactory.create(
            priority=p3_priority,
            _status=IncidentStatus.INVESTIGATING,  # Test from INVESTIGATING
        )

        # Create form with the incident
        form = UpdateStatusForm(incident=incident)

        # Check that CLOSED is in the choices from INVESTIGATING status for P3+ but POST_MORTEM is NOT
        status_choices = dict(form.fields["status"].choices)
        available_statuses = self._get_available_statuses(status_choices)
        assert IncidentStatus.CLOSED in available_statuses
        assert IncidentStatus.POST_MORTEM not in available_statuses, "P3+ incidents should not have post-mortem status available"

    def test_p1_p2_can_go_from_mitigated_to_postmortem(self):
        """P1/P2 incidents can go from Mitigated to Post-mortem."""

        # Get existing P1 or P2 priority that needs post-mortem
        p1_priority = Priority.objects.filter(
            value__in=[1, 2], needs_postmortem=True
        ).first()

        if not p1_priority:
            pytest.skip("No P1/P2 priority with needs_postmortem=True found in database")

        # Get PRD environment
        prd_env = Environment.objects.filter(value="PRD").first()
        if not prd_env:
            pytest.skip("No PRD environment found in database")

        # Create an incident with P1/P2 priority in Mitigated status with PRD env
        incident = IncidentFactory.create(
            priority=p1_priority,
            environment=prd_env,
            _status=IncidentStatus.MITIGATED,  # MITIGATED = "Mitigated"
        )

        # Create form with the incident
        form = UpdateStatusForm(incident=incident)

        # Check that POST_MORTEM is available but CLOSED is not
        status_choices = dict(form.fields["status"].choices)
        available_statuses = self._get_available_statuses(status_choices)
        assert IncidentStatus.POST_MORTEM in available_statuses, "P1/P2 should be able to go to Post-mortem from Mitigated"
        assert IncidentStatus.CLOSED not in available_statuses, "P1/P2 should NOT be able to skip Post-mortem"

    def test_complete_workflow_transitions_p3_plus(self):
        """Test the correct workflow for P3+: can close with reason from early statuses, normal close from Mitigated."""
        # Get P3+ priority
        p3_priority = Priority.objects.filter(
            value__gte=3, needs_postmortem=False
        ).first()

        if not p3_priority:
            pytest.skip("No P3+ priority found")

        # Test workflow for P3+
        test_cases = [
            (IncidentStatus.OPEN, True),           # Can close with reason
            (IncidentStatus.INVESTIGATING, True),  # Can close with reason
            (IncidentStatus.MITIGATING, False),        # Cannot close from Mitigating (must go to Mitigated first)
            (IncidentStatus.MITIGATED, True),          # Can close normally from Mitigated
        ]

        for current_status, should_have_closed in test_cases:
            incident = IncidentFactory.create(
                priority=p3_priority,
                _status=current_status,
            )

            form = UpdateStatusForm(incident=incident)
            status_choices = dict(form.fields["status"].choices)
            available_statuses = self._get_available_statuses(status_choices)

            if should_have_closed:
                assert IncidentStatus.CLOSED in available_statuses, f"P3+ should be able to go to Closed from {current_status.label}"
            else:
                assert IncidentStatus.CLOSED not in available_statuses, f"P3+ should NOT be able to go to Closed from {current_status.label}"

            # P3+ should NEVER have post-mortem available
            assert IncidentStatus.POST_MORTEM not in available_statuses, f"P3+ should NEVER have post-mortem available from {current_status.label}"

    def test_closure_reason_required_from_early_statuses(self):
        """Test that closure reason is required when closing from Opened or Investigating."""
        # Test for both P1/P2 and P3+ priorities
        priorities = [
            Priority.objects.filter(value__in=[1, 2]).first(),  # P1/P2
            Priority.objects.filter(value__gte=3).first(),       # P3+
        ]

        for priority in priorities:
            if not priority:
                continue

            # Test from Opened and Investigating
            for status in [IncidentStatus.OPEN, IncidentStatus.INVESTIGATING]:
                incident = IncidentFactory.create(
                    priority=priority,
                    _status=status,
                )

                # Check that requires_closure_reason returns True
                assert UpdateStatusForm.requires_closure_reason(
                    incident, IncidentStatus.CLOSED
                ), f"Should require closure reason from {status.label} for {priority.name}"

            # Test from other statuses - should NOT require reason
            for status in [IncidentStatus.MITIGATING, IncidentStatus.MITIGATED, IncidentStatus.POST_MORTEM]:
                incident = IncidentFactory.create(
                    priority=priority,
                    _status=status,
                )

                # Check that requires_closure_reason returns False
                assert not UpdateStatusForm.requires_closure_reason(
                    incident, IncidentStatus.CLOSED
                ), f"Should NOT require closure reason from {status.label} for {priority.name}"

    def test_cannot_close_from_mitigating(self):
        """Test that incidents cannot be closed directly from Mitigating status."""
        # Test for all priority levels
        priorities = Priority.objects.all()

        for priority in priorities:
            incident = IncidentFactory.create(
                priority=priority,
                _status=IncidentStatus.MITIGATING,  # Mitigating status
            )

            form = UpdateStatusForm(incident=incident)
            status_choices = dict(form.fields["status"].choices)
            available_statuses = self._get_available_statuses(status_choices)

            # Should NOT have Closed option from Mitigating
            assert IncidentStatus.CLOSED not in available_statuses, f"Should NOT be able to close from Mitigating for {priority.name}"

            # Should have Post-mortem option only for P1/P2 priorities that need postmortem
            if priority.needs_postmortem:
                # For P1/P2, should still NOT have post-mortem from MITIGATING (Mitigating)
                assert IncidentStatus.POST_MORTEM not in available_statuses, f"P1/P2 should NOT be able to go to Post-mortem from Mitigating for {priority.name}"
            else:
                # For P3+, should not have post-mortem at all
                assert IncidentStatus.POST_MORTEM not in available_statuses, f"P3+ should not have post-mortem available for {priority.name}"

    def test_no_incident_shows_all_statuses(self):
        """When no incident is provided, all statuses should be available."""
        # Create form without incident
        form = UpdateStatusForm()

        # Check that all statuses including CLOSED are in the choices
        status_choices = dict(form.fields["status"].choices)
        available_statuses = self._get_available_statuses(status_choices)
        assert IncidentStatus.CLOSED in available_statuses
        assert IncidentStatus.POST_MORTEM in available_statuses

    def test_requires_closure_reason_non_closed_status(self):
        """Test requires_closure_reason with non-CLOSED target status."""
        incident = IncidentFactory.create(_status=IncidentStatus.OPEN)

        # Should not require reason for non-CLOSED statuses
        assert not UpdateStatusForm.requires_closure_reason(incident, IncidentStatus.INVESTIGATING)
        assert not UpdateStatusForm.requires_closure_reason(incident, IncidentStatus.MITIGATING)
        assert not UpdateStatusForm.requires_closure_reason(incident, IncidentStatus.MITIGATED)
        assert not UpdateStatusForm.requires_closure_reason(incident, IncidentStatus.POST_MORTEM)

    def test_incident_status_edge_cases(self):
        """Test edge cases for incident status transitions."""
        # Test with incident in default fallback case
        p1_priority = Priority.objects.filter(
            value__in=[1, 2], needs_postmortem=True
        ).first()

        if not p1_priority:
            pytest.skip("No P1/P2 priority found")

        prd_env = Environment.objects.filter(value="PRD").first()
        if not prd_env:
            pytest.skip("No PRD environment found")

        # Create incident with an undefined status (should hit default case)
        incident = IncidentFactory.create(
            priority=p1_priority,
            environment=prd_env,
            _status=IncidentStatus.CLOSED  # Already closed
        )

        form = UpdateStatusForm(incident=incident)
        status_choices = dict(form.fields["status"].choices)
        available_statuses = self._get_available_statuses(status_choices)

        # Should use default choices_lte for unknown/closed status
        assert IncidentStatus.CLOSED in available_statuses



from __future__ import annotations

import logging
from typing import Any

from django.core.management.base import BaseCommand, CommandParser

from firefighter.confluence.tasks.sort_runbooks import sort_runbooks

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Sort runbooks."
    requires_migrations_checks = False
    requires_system_checks: list[str] = []

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument("--dry-run", dest="dry_run", action="store_true")
        parser.set_defaults(dry_run=False)

    def handle(self, *args: Any, **options: Any) -> Any:
        return sort_runbooks(dry_run=options["dry_run"])



from __future__ import annotations

import logging
import uuid
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any

from django.db import models
from django_stubs_ext.db.models import TypedModelMeta
from slack_sdk.errors import SlackApiError

from firefighter.incidents.models import IncidentUpdate
from firefighter.incidents.models.incident import Incident
from firefighter.slack.models.conversation import Conversation
from firefighter.slack.models.user import SlackUser
from firefighter.slack.slack_app import DefaultWebClient, SlackApp, slack_client

if TYPE_CHECKING:
    from slack_sdk.web.client import WebClient
    from slack_sdk.web.slack_response import SlackResponse

logger = logging.getLogger(__name__)


class MessageManager(models.Manager["Message"]):
    def get_or_none(self, **kwargs: Any) -> Message | None:
        try:
            return self.get(**kwargs)
        except Message.DoesNotExist:
            return None

    def create_from_slack_response(
        self, response: SlackResponse, **kwargs: Any
    ) -> Message | None:
        msg_res = response["message"]
        if msg_res is None:
            raise ValueError("msg_res is None")
        user = SlackUser.objects.get_user_by_slack_id(slack_id=msg_res["user"])
        if user is None:
            logger.warning(f"User not found for slack_id {msg_res['user']}")
            return None
        if not hasattr(user, "slack_user") or user.slack_user is None:
            logger.warning(f"User {user.id} has no slack_user")
            return None

        conversation = Conversation.objects.get(channel_id=response["channel"])
        ts = datetime.fromtimestamp(float(msg_res["ts"]), tz=UTC)
        return self.create(
            conversation=conversation,
            ts=ts,
            type=msg_res["type"],
            blocks=msg_res["blocks"],
            user=user.slack_user,
            text=msg_res["text"],
            **kwargs,
        )


class Message(models.Model):
    """Model a Slack API Conversation.
    A Slack Conversation can be a public channel, a private channel, a direct message, or a multi-person direct message.
    Reference: https://api.slack.com/types/conversation.
    """

    objects: MessageManager = MessageManager()

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    # Conversation ID + Timestamp is the True composite PK (unsupported by Django)
    conversation = models.ForeignKey(
        Conversation, on_delete=models.CASCADE, db_index=True
    )
    ts = models.DateTimeField(
        primary_key=False, help_text="UTC Timestamp of the message"
    )

    user = models.ForeignKey(SlackUser, on_delete=models.CASCADE)

    thread_ts = models.DateTimeField(null=True, blank=True)

    type = models.CharField(max_length=80)
    subtype = models.CharField(max_length=80, blank=True)

    text = models.TextField(max_length=40000, blank=True)
    blocks = models.JSONField(null=True, blank=True)

    incident_update = models.ForeignKey(
        IncidentUpdate,
        on_delete=models.CASCADE,
        related_name="slack_message",
        null=True,
    )

    incident = models.ForeignKey(
        Incident,
        on_delete=models.CASCADE,
        related_name="slack_message",
        null=True,
    )

    ff_type = models.CharField(max_length=64, blank=True)
    if TYPE_CHECKING:
        user_id: uuid.UUID | None
        conversation_id: uuid.UUID

    class Meta(TypedModelMeta):
        constraints = [
            models.UniqueConstraint(
                fields=["ts", "conversation"], name="message_ts_unicity"
            )
        ]

    def __str__(self) -> str:
        return f"{self.ts}@{self.conversation.name}: {self.text} {self!r}"

    def __repr__(self) -> str:
        return f"{self.ts}@{self.conversation_id}: {self.text}"

    def get_absolute_url(self) -> str:
        return self.get_permalink

    @property
    def get_permalink(self) -> str:
        return f"{SlackApp().details['url']}archives/{self.conversation.channel_id}/p{str(self.ts.timestamp()).replace('.', '')}"

    @slack_client
    def delete_self_and_slack(self, client: WebClient = DefaultWebClient) -> None:
        channel_id = self.conversation.channel_id
        if channel_id is None:
            raise ValueError("channel_id is None. Cannot delete message.")
        try:
            res = client.chat_delete(channel=channel_id, ts=str(self.ts.timestamp()))
            if res.get("ok"):
                self.delete()
            else:
                logger.warning(f"Failed to delete message {self} from Slack: {res}")
        except SlackApiError as e:
            logger.warning(f"Failed to delete message {self} from Slack: {e}")
